import os
from pathlib import Path

from omegaconf import OmegaConf

ROOT_DIR = Path(os.path.abspath(__file__)).parent.parent
if 'DATA_DIR' not in os.environ:
    DATA_DIR = ROOT_DIR / "data"
else:
    DATA_DIR = Path(os.environ['DATA_DIR'])

assert DATA_DIR.exists(), f"{DATA_DIR} does not exist"
DEFAULT_CONFIG = OmegaConf.load(ROOT_DIR / "grams.yaml")



from pathlib import Path

from sm.misc import RocksDBStore
from typing import Optional, Union


class Wikipedia2WikidataDB(RocksDBStore[str, str]):
    instance = None

    @staticmethod
    def get_instance(dbfile: Optional[Union[Path, str]] = None, read_only: bool = False):
        if Wikipedia2WikidataDB.instance is None:
            Wikipedia2WikidataDB.instance = Wikipedia2WikidataDB(dbfile, read_only=read_only)
        return Wikipedia2WikidataDB.instance

    def deserialize(self, value):
        return value.decode()





import glob
from collections import defaultdict
from pathlib import Path
from typing import List, Dict, Tuple, Callable, Any, Optional
import grams.misc as M
from grams.config import ROOT_DIR


def clean_table_linker(infile, outfile):
    rows = M.deserialize_csv(infile)
    cname2idx = {v: k for k, v in enumerate(rows[0])}

    cells = defaultdict(dict)
    for row in rows[1:]:
        ci = int(row[cname2idx['column']])
        ri = int(row[cname2idx['row']])
        gt_ent = row[cname2idx['GT_kg_id']]
        pred_ent = row[cname2idx['kg_id']]
        score = row[cname2idx['siamese_pred']]

        if (ri, ci) in cells:
            assert cells[ri, ci]['gt'] == gt_ent
        cells[ri, ci]['gt'] = gt_ent
        if 'pred_ent' not in cells[ri, ci]:
            cells[ri, ci]['pred_ent'] = []
        cells[ri, ci]['pred_ent'].append(f"{pred_ent}:{score}")

    links = []
    for (ri, ci), o in cells.items():
        links.append([ri, ci, o['gt']] + o['pred_ent'])
    M.serialize_csv(links, outfile, delimiter="\t")


if __name__ == '__main__':
    for infile in glob.glob(str(ROOT_DIR / "examples/t2dv2/tables/*.table_linker.csv")):
        infile = Path(infile)
        outfile = infile.parent / f"{infile.name.split('.')[0]}.candidates.tsv"
        clean_table_linker(infile, outfile)


import copy
from dataclasses import dataclass
from operator import attrgetter, itemgetter

import networkx as nx
import matplotlib.pyplot as plt
from typing import Set, List, Dict, Tuple, Callable, FrozenSet, Optional, Any

"""
Find a top-k approximate minimum steiner arborescence using the BANK algorithm
"""


@dataclass
class Edge:
    __slots__ = ("id", "source_id", "target_id", "edge_key", "weight", "n_edges")
    id: str
    source_id: str
    target_id: str
    edge_key: str
    weight: float
    n_edges: int


@dataclass
class Solution:
    id: FrozenSet[str]
    graph: nx.MultiDiGraph
    weight: float

    def get_n_edges(self):
        if not hasattr(self, "n_edges"):
            self.n_edges = sum(e.n_edges for _, _, e in self.graph.edges(data='data'))
        return self.n_edges


class NoSingleRootException(Exception):
    pass


# noinspection PyMethodMayBeStatic
class SteinerTreeBankSolver:

    def __init__(self,
                 original_graph: nx.MultiDiGraph,
                 terminal_nodes: Set[str],
                 weight_fn: Callable[[dict], float],
                 solution_cmp_fn: Optional[Callable[[Solution], Any]]=None,
                 top_k_st: int = 10,
                 top_k_path: int = 10,
                 allow_shorten_graph: bool = True):
        # original graph
        self.original_graph = original_graph
        # function that extract weights
        self.weight_fn = weight_fn
        # function to compare & sort solutions
        self.solution_cmp_fn = solution_cmp_fn or attrgetter('weight')
        # graph that the bank algorithm will work with
        self.graph: nx.MultiDiGraph = None
        # output graphs
        self.solutions: List[Solution] = []
        self.terminal_nodes = terminal_nodes
        self.top_k_st = top_k_st
        self.top_k_path = top_k_path
        self.allow_shorten_graph = allow_shorten_graph

    def run(self):
        self.graph = self._preprocessing(self.original_graph, self.weight_fn)

        if nx.is_weakly_connected(self.graph):
            self.solutions = self._solve(self.graph, self.terminal_nodes, self.top_k_st, self.top_k_path)
        else:
            graphs = self._split_connected_components(self.graph)
            final_solutions = None
            for g in graphs:
                terminal_nodes = self.terminal_nodes.intersection(list(g.nodes))
                solutions = self._solve(g, terminal_nodes, self.top_k_st, self.top_k_path)
                if final_solutions is None:
                    final_solutions = solutions
                else:
                    next_solutions = []
                    for current_sol in final_solutions:
                        for sol in solutions:
                            next_solutions.append(self._merge_graph(current_sol.graph, sol.graph))
                    final_solutions = self._sort_solutions(next_solutions, return_solution=True)[:self.top_k_st]

            self.solutions = final_solutions

        # [self._get_roots(sol.graph) for sol in self.solutions]
        # [sol.weight for sol in self.solutions]
        return [self._postprocessing(self.original_graph, self.graph, sol.graph) for sol in self.solutions], self.solutions

    def _preprocessing(self, g: nx.MultiDiGraph, weight_fn: Callable[[dict], float]):
        ng = nx.MultiDiGraph()

        # convert edges
        edge_id_count = 0
        for uid, vid, eid, edata in g.edges(data=True, keys=True):
            ng.add_edge(uid, vid, key=eid, data=Edge(id=str(edge_id_count), source_id=uid, target_id=vid, edge_key=eid,
                                                     weight=weight_fn(edata), n_edges=1))
            edge_id_count += 1

        if self.allow_shorten_graph:
            # shorten path using the following heuristic
            # for a node that only connect two nodes (indegree & outdegree = 1) and not terminal nodes, we replace the node by one edge
            # map from the replaced edge to the removed nodes
            removed_nodes = {}
            for uid in list(ng.nodes):
                if uid not in self.terminal_nodes and g.in_degree(uid) == 1 and g.out_degree(uid) == 1:
                    inedge: Edge = next(iter(ng.in_edges(uid, data='data')))[-1]
                    outedge: Edge = next(iter(ng.out_edges(uid, data='data')))[-1]
                    if inedge.edge_key == outedge.edge_key:
                        new_edge_key = inedge.edge_key
                    else:
                        new_edge_key = f"{inedge.edge_key}-{outedge.edge_key}"
                    if not ng.has_edge(inedge.source_id, outedge.target_id, key=new_edge_key):
                        # replace it only if we don't have that link before
                        ng.remove_node(uid)
                        ng.add_edge(inedge.source_id, outedge.target_id, key=new_edge_key, data=Edge(id=str(edge_id_count),
                                                                                                    source_id=inedge.source_id,
                                                                                                    target_id=outedge.target_id,
                                                                                                    edge_key=new_edge_key,
                                                                                                    weight=inedge.weight + outedge.weight,
                                                                                                    n_edges=2))
                        removed_nodes[str(edge_id_count)] = (uid, inedge, outedge)
                        edge_id_count += 1

            # store the removed nodes to restore it later
            ng.graph['removed_nodes'] = removed_nodes
        return ng

    def _postprocessing(self, origin_graph: nx.MultiDiGraph, prep_graph: nx.MultiDiGraph, out_graph: nx.MultiDiGraph):
        if self.allow_shorten_graph:
            removed_nodes = prep_graph.graph['removed_nodes']
        else:
            removed_nodes = set()

        g = origin_graph.copy()
        selected_edges = set()
        for uid, vid, edge in out_graph.edges(data='data'):
            if edge.id in removed_nodes:
                removed_node, inedge, outedge = removed_nodes[edge.id]
                selected_edges.add((inedge.source_id, inedge.target_id, inedge.edge_key))
                selected_edges.add((outedge.source_id, outedge.target_id, outedge.edge_key))
            else:
                selected_edges.add((uid, vid, edge.edge_key))

        for uid, vid, eid in list(g.edges(keys=True)):
            if (uid, vid, eid) not in selected_edges:
                g.remove_edge(uid, vid, eid)
        for uid in list(g.nodes):
            if g.in_degree(uid) == 0 and g.out_degree(uid) == 0:
                g.remove_node(uid)
        return g

    def _merge_graph(self, g1: nx.MultiDiGraph, g2: nx.MultiDiGraph) -> nx.MultiDiGraph:
        g = g1.copy()
        for uid, vid, eid, edata in g2.edges(keys=True, data=True):
            g.add_edge(uid, vid, key=eid, **edata)
        return g

    def _split_connected_components(self, g: nx.MultiDiGraph):
        connected_comps = [
            comp
            for comp in nx.weakly_connected_components(g)
            # must have at least two terminal nodes (to form a graph)
            if len(self.terminal_nodes.intersection(comp)) > 1
        ]

        node2comp = {}
        for i, comp in enumerate(connected_comps):
            for uid in comp:
                node2comp[uid] = i

        subgs = [nx.MultiDiGraph() for _ in range(len(connected_comps))]
        for uid, vid, eid, edata in g.edges(keys=True, data=True):
            if uid not in node2comp:
                continue
            subgs[node2comp[uid]].add_edge(uid, vid, key=eid, **edata)
        return subgs

    def _solve(self, g: nx.MultiDiGraph, terminal_nodes: Set[str], top_k_st: int, top_k_path: int):
        """Despite the name, this is finding steiner tree. Assuming their is a root node that connects all
        terminal nodes together.
        """
        roots = set(g.nodes.keys())

        attr_visit_hists = []
        # to ensure the order
        for uid in list(sorted(terminal_nodes)):
            visit_hist = UpwardTraversal.top_k_beamsearch(g, uid, top_k_path)
            roots = roots.intersection(visit_hist.paths.keys())
            attr_visit_hists.append((uid, visit_hist))

        if len(roots) == 0:
            # there is no nodes that can connect to all terminal nodes either this are disconnected
            # components or you pass a directed graph with weakly connected components (a -> b <- c)
            if nx.is_weakly_connected(g):
                # perhaps, we can break the weakly connected components by breaking one of the link (a -> b <- c)
                raise NoSingleRootException(
                    "You pass a weakly connected component and there are parts of the graph like this (a -> b <- c). Fix it before running this algorithm")
            raise Exception("Your graph is disconnected. Consider splitting them before calling bank solver")

        # to ensure the order again & remove randomness
        roots = sorted(roots)

        # merge the paths using beam search
        results = []
        for root in roots:
            current_states = []
            uid, visit_hist = attr_visit_hists[0]
            for path in visit_hist.paths[root]:
                pg = nx.MultiDiGraph()
                if len(path.path) > 0:
                    assert uid == path.path[0].target_id
                pg.add_node(uid)
                for e in path.path:
                    pg.add_node(e.source_id)
                    pg.add_edge(e.source_id, e.target_id, e.edge_key, weight=e.weight, data=e)
                current_states.append(pg)

            if len(current_states) > top_k_st:
                current_states = self._sort_solutions(current_states)[:top_k_st]

            for uid, visit_hist in attr_visit_hists[1:]:
                next_states = []
                for state in current_states:
                    for path in visit_hist.paths[root]:
                        pg = state.copy()
                        if len(path.path) > 0:
                            assert uid == path.path[0].target_id
                        pg.add_node(uid)
                        for e in path.path:
                            if e.source_id not in pg.nodes:
                                pg.add_node(e.source_id)
                            # TODO: here we don't check by edge_key because we may create another edge of different key
                            # hope this new path has been exploited before.
                            if not pg.has_edge(e.source_id, e.target_id):
                                pg.add_edge(e.source_id, e.target_id, key=e.edge_key, weight=e.weight, data=e)

                        # after add a path to the graph, it can create new cycle, detect and fix it
                        try:
                            # note that the function I'm using doesn't return parallel edges
                            cycles_iter = nx.find_cycle(pg, uid, orientation='reverse')
                            # we can show that if the graph contain cycle, there is a better path
                            #
                            # cycles_iter = [(uid, vid) for uid, vid, eid, orien in cycles_iter]
                            # for _g in self._break_cycles(root, pg, cycles_iter):
                            #     next_states.append(_g)
                        except nx.NetworkXNoCycle:
                            next_states.append(pg)
                        if len(list(pg.edges)) != len({(uid, vid) for uid, vid, eid in pg.edges(keys=True)}):
                            assert False
                if len(next_states) > top_k_st:
                    next_states = self._sort_solutions(next_states)[:top_k_st]
                current_states = next_states
                # cgs = [g for g in next_states if len(list(nx.simple_cycles(g))) > 0]
                # nx.draw_networkx(cgs[0]); plt.show()
                # nx.draw(cgs[0]); plt.show()
            results += current_states
        return self._sort_solutions(results, return_solution=True)

    def _break_cycles(self, root: str, g: nx.MultiDiGraph, cycles_iter: List[Tuple[str, str]]):
        # g = current_states[0]; g = self.output_graphs[0]
        # pos = nx.kamada_kawai_layout(g); nx.draw_networkx(g, pos); nx.draw_networkx_edge_labels(g, pos, edge_labels={(u, v): d for u, v, d in g.edges(keys=True)}); plt.show()
        # nx.draw(g); plt.show()
        return self._break_cycles_brute_force(root, g, cycles_iter)

    def _break_cycles_brute_force(self, root: str, g: nx.MultiDiGraph, cycles_iter: List[Tuple[str, str]]):
        # one side effect of this approach is that it may separate the graph
        # currently we say it only happen when the target node of the remove edge only has one incoming edge
        # if it has two edges, then the other source (not of the remove edge) must have a path from root -> itself
        # now, since we have cascade removing, if the path doesn't go through the removed node, then it's okay,
        # if the path goes through the remove node, it's impossible since we only remove node that indegree == 0 or
        # outdegree == 0
        parallel_edges = []
        for uid, vid in cycles_iter:
            if g.in_degree(vid) == 1:
                # we can't remove this edge, as removing it will make it's unreachable from the root
                # so just skip edge
                continue
            edges = [edata['data'] for eid, edata in g[uid][vid].items()]
            edge_weight = min(edges, key=attrgetter('weight')).weight
            parallel_edges.append((uid, vid, edges, edge_weight))

        # not comparing based on weight anymore since psl can be quite difficult to select correct edge
        # min_edge_weight = min(parallel_edges, key=itemgetter(3))[3]
        # unbreakable = {i for i, x in enumerate(parallel_edges) if x[3] == min_edge_weight}
        # new_graphs = []
        # if len(unbreakable) < len(parallel_edges):
        #     # it's great! we have some edge to break!
        #     # for each breakable edge, we will have a new graph
        #     for i, item in enumerate(parallel_edges):
        #         if i in unbreakable:
        #             continue
        #         ng: nx.MultiDiGraph = g.copy()
        #         ng.remove_edge(item[0], item[1])
        #         ng = self._remove_redundant_nodes(root, ng)
        #         new_graphs.append(ng)
        # else:
            # so bad, we have to try one by one
        new_graphs = []
        for uid, vid, edges, edge_weight in parallel_edges:
            ng: nx.MultiDiGraph = g.copy()
            ng.remove_edge(uid, vid)
            # self._draw(ng); self._draw(g)
            ng = self._remove_redundant_nodes(root, ng)
            new_graphs.append(ng)

        # just assert if it works as expected
        for ng in new_graphs:
            assert len(list(nx.simple_cycles(ng))) == 0
        return new_graphs

    def _sort_solutions(self, graphs: List[nx.MultiDiGraph], return_solution: bool = False):
        """Sort the solution by th"""
        solutions = {}
        for g in graphs:
            id = frozenset((e.id for _, _, e in g.edges(data='data')))
            if id in solutions:
                continue

            weight = sum(e.weight for _, _, e in g.edges(data='data'))
            solutions[id] = Solution(id, g, weight)

        solutions = sorted(solutions.values(), key=self.solution_cmp_fn)
        if return_solution:
            return solutions
        return [sol.graph for sol in solutions]

    def _remove_redundant_nodes(self, root: str, g: nx.MultiDiGraph):
        # remove nodes in the graph that shouldn't be in a steiner tree
        while True:
            removed_nodes = []
            for uid in g.nodes:
                if uid == root or uid in self.terminal_nodes:
                    continue
                if g.in_degree(uid) == 0 or g.out_degree(uid) == 0:
                    removed_nodes.append(uid)
            if len(removed_nodes) == 0:
                break
            for uid in removed_nodes:
                g.remove_node(uid)
        return g

    def _get_roots(self, g: nx.MultiDiGraph):
        """This function is mainly used for debugging"""
        return [uid for uid in g if g.in_degree(uid) == 0]

    @staticmethod
    def _draw(g: nx.MultiDiGraph):
        """This function is mainly used for debugging"""
        pos = nx.kamada_kawai_layout(g)
        nx.draw_networkx(g, pos)
        nx.draw_networkx_edge_labels(g, pos, edge_labels={(u, v): d for u, v, d in g.edges(keys=True)})
        plt.show()


class CycleBreaker:
    def spanning_arborescence(self, g: nx.MultiDiGraph, root: str, terminal_nodes: Set[str]):
        resp = nx.algorithms.tree.branchings.minimum_spanning_arborescence(g, attr='weight', default=None,
                                                                           preserve_attrs=True)
        ng = nx.MultiDiGraph()
        for uid, vid, edge in resp.edges(data='data'):
            ng.add_edge(uid, vid, key=edge.edge_key, weight=edge.weight, data=edge)
        # after breaking we may have multiple root nodes or leaf so we just need to clean them
        return [self._truncate_redundant_nodes(ng, root, terminal_nodes)]

    def one_edge_at_a_time(self, og: nx.MultiDiGraph, root: str, terminal_nodes: Set[str]):
        """This approach doesn't allow us to select the graph that has shorter length due to multiple cycles in the graph"""
        current_graphs = [og]
        has_cycle = True
        while has_cycle:
            # loop until we don't have any cycle left
            new_graphs = []
            for g in current_graphs:
                for cycle in nx.simple_cycles(g):
                    cycle.append(cycle[0])
                    edge_groups = []
                    for i in range(len(cycle) - 1):
                        uid = cycle[i]
                        vid = cycle[i+1]
                        # we may have more than one edge between two nodes
                        edges = [edata['data'] for eid, edata in g[uid][vid].items()]
                        edge_weight = min(edges, key=attrgetter('weight'))
                        edge_groups.append((uid, vid, edges, edge_weight))

                    min_edge_weight = min(edge_groups, itemgetter(3))
                    unbreakable = {i for i, x in enumerate(edge_groups) if x[3] == min_edge_weight}
                    if len(unbreakable) < len(edge_groups):
                        # it's great! we have some edge to break!
                        # for each breakable edge, we will have a new graph
                        for i, item in enumerate(edge_groups):
                            if i in unbreakable:
                                continue
                            ng: nx.MultiDiGraph = g.copy()
                            # we remove all instead of just one
                            ng.remove_edge(item[0], item[1])
                            ng = self._truncate_redundant_nodes(ng, root, terminal_nodes)
                            new_graphs.append(ng)
                    else:
                        # so bad, we have to try one by one
                        tmp = []
                        for uid, vid, edges, edge_weight in edge_groups:
                            ng: nx.MultiDiGraph = g.copy()
                            ng.remove_edge(uid, vid)
                            ng = self._truncate_redundant_nodes(ng, root, terminal_nodes)
                            # given a score for each split, we prefer the one that is smaller?
                            # however, we can't calculate the depth since the graph may contain multiple cycles
                            # this make this approach not good.
                            tmp.append(ng)

                        # find which one is better

                    pass

    def _truncate_redundant_nodes(self, g: nx.MultiDiGraph, root: str, terminal_nodes: Set[str]):
        while True:
            removed_nodes = []
            for uid in g.nodes:
                if uid == root or uid in terminal_nodes:
                    continue
                if g.in_degree(uid) == 0 or g.out_degree(uid) == 0:
                    removed_nodes.append(uid)
            if len(removed_nodes) == 0:
                break
            for uid in removed_nodes:
                g.remove_node(uid)
        return g


@dataclass
class UpwardPath:
    __slots__ = ("visited_nodes", "path", "weight")

    visited_nodes: Set[str]
    path: List[Edge]
    weight: float

    @staticmethod
    def empty(start_node_id: str):
        return UpwardPath({start_node_id}, [], 0.0)

    def push(self, edge: Edge) -> 'UpwardPath':
        c = self.clone()
        c.path.append(edge)
        c.visited_nodes.add(edge.source_id)
        c.weight += edge.weight
        return c

    def clone(self):
        return UpwardPath(copy.copy(self.visited_nodes), copy.copy(self.path), self.weight)


@dataclass
class UpwardTraversal:
    __slots__ = ("source_id", "paths")

    source_id: str

    # storing that we can reach this node through those list of paths
    paths: Dict[str, List[UpwardPath]]

    @staticmethod
    def top_k_beamsearch(g: nx.MultiDiGraph, start_node_id: str, top_k_path: int):
        travel_hist = UpwardTraversal(start_node_id, dict())
        travel_hist.paths[start_node_id] = [UpwardPath.empty(start_node_id)]
        for source_id, target_id, edge_id, orientation in nx.edge_bfs(g, start_node_id, orientation='reverse'):
            if source_id not in travel_hist.paths:
                travel_hist.paths[source_id] = []

            edge: Edge = g.edges[source_id, target_id, edge_id]['data']
            #
            # for edata in g[source_id][target_id].values():
            #     edge: Edge = edata['data']
            for path in travel_hist.paths[target_id]:
                if source_id in path.visited_nodes:
                    # path will become loopy, which we don't want to do
                    continue
                path = path.push(edge)
                travel_hist.paths[source_id].append(path)

            # we trim the number of paths in here
            if len(travel_hist.paths[source_id]) > top_k_path:
                # calculate the score of each path, and then select the best one
                travel_hist.sort_paths(source_id)
                travel_hist.paths[source_id] = travel_hist.paths[source_id][:top_k_path]
        return travel_hist

    def sort_paths(self, node_id: str):
        self.paths[node_id] = sorted(self.paths[node_id], key=attrgetter('weight'))



from dataclasses import dataclass
import os
from operator import itemgetter
from pathlib import Path
from typing import List, Dict, Tuple, Callable, Any, Optional, Set, Iterable, Union

import networkx as nx
import requests
from omegaconf import OmegaConf
from rdflib import RDFS

import sm.misc as M
import sm.outputs as O
import grams.inputs as I

from grams.algorithm.data_graph import build_data_graph, BuildDGOption
from grams.algorithm.kg_index import TraversalOption, KGObjectIndex
from grams.algorithm.semantic_graph import SemanticGraphConstructor, SemanticGraphConstructorArgs, viz_sg
from grams.algorithm.sm_wikidata import WikidataSemanticModelHelper
from grams.config import DEFAULT_CONFIG, ROOT_DIR

from grams.algorithm.psl_solver import PSLSteinerTreeSolver

from kgdata.wikidata.models import QNode, WDProperty, WDClass, WDQuantityPropertyStats
from kgdata.wikidata.db import get_qnode_db, get_wdprop_db, get_wdclass_db, query_wikidata_entities


@dataclass
class Annotation:
    sm: O.SemanticModel
    dg: nx.MultiDiGraph
    sg: nx.MultiDiGraph
    pred_sg: nx.MultiDiGraph
    pred_cta: Dict[int, Dict[str, float]]


class GRAMS:
    def __init__(self, data_dir: str, cfg=None, proxy: bool=True):
        self.timer = M.Timer()
        self.cfg = cfg if cfg is None else DEFAULT_CONFIG
        self.is_proxy_db = proxy

        with self.timer.watch('init grams db'):
            read_only = not proxy
            self.qnodes = get_qnode_db(os.path.join(data_dir, "qnodes.db"), compression=True, read_only=read_only, proxy=proxy, is_singleton=True)
            self.wdclasses = get_wdclass_db(os.path.join(data_dir, "wdclasses.db"), compression=False, read_only=read_only, proxy=proxy, is_singleton=True)
            self.wdprops = get_wdprop_db(os.path.join(data_dir, "wdprops.db"), compression=False, read_only=read_only, proxy=proxy, is_singleton=True)
            self.wd_numprop_stats = WDQuantityPropertyStats.from_dir(os.path.join(data_dir, "quantity_prop_stats"))

        self.build_dg_option = getattr(BuildDGOption, cfg.data_graph.options[0])
        for op in cfg.data_graph.options[1:]:
            self.build_dg_option = self.build_dg_option | getattr(BuildDGOption, op)

    def annotate(self, table: I.LinkedTable, verbose: bool=False):
        qnode_ids = {link.qnode_id
                     for rlinks in table.links for links in rlinks
                     for link in links if link.qnode_id is not None}
        if table.context.page_qnode is not None:
            qnode_ids.add(table.context.page_qnode)

        with self.timer.watch('retrieving qnodes'):
            qnodes = self.get_entities(qnode_ids, n_hop=2, verbose=verbose)
        wdclasses = self.wdclasses.cache_dict()
        wdprops = self.wdprops.cache_dict()

        with self.timer.watch('build kg object index'):
            kg_object_index = KGObjectIndex.from_qnodes(
                qnode_ids, qnodes, wdprops,
                n_hop=self.cfg.data_graph.max_n_hop, traversal_option=TraversalOption.TransitiveOnly)

        with self.timer.watch("build dg & sg"):
            dg = build_data_graph(table, qnodes, wdprops,
                                  kg_object_index, max_n_hop=self.cfg.data_graph.max_n_hop,
                                  options=self.build_dg_option)
            constructor = SemanticGraphConstructor([
                SemanticGraphConstructor.init_sg,
            ], qnodes, wdclasses, wdprops)
            sg = constructor.run(table, dg, debug=False).sg
        
        with self.timer.watch('run inference'):
            psl_solver = PSLSteinerTreeSolver(
                qnodes, wdclasses, wdprops, self.wd_numprop_stats,
                disable_rules=set(self.cfg.psl.disable_rules), sim_fn=None,
                # cache_dir=outdir / f"{override_psl_cachedir}cache/psl",
                postprocessing_method=self.cfg.psl.postprocessing, enable_logging=self.cfg.psl.enable_logging)
            pred_sg, pred_cta = psl_solver.run(dict(table=table, semanticgraph=sg, datagraph=dg))
            pred_cta = {int(ci.replace("column-", "")): classes for ci, classes in pred_cta.items()}
            cta = {ci: max(classes.items(), key=itemgetter(1))[0] for ci, classes in pred_cta.items()}

        sm = self.create_sm_from_cta_cpa(table, pred_sg, cta, qnodes, wdclasses, wdprops)
        return Annotation(sm=sm, dg=dg, sg=sg, pred_sg=pred_sg, pred_cta=cta)

    def create_sm_from_cta_cpa(self, table: I.LinkedTable, sg: nx.MultiDiGraph, cta: Dict[int, str], qnodes: Dict[str, QNode], wdclasses: Dict[str, WDClass], wdprops: Dict[str, WDProperty]):
        sm = O.SemanticModel()
        sm_helper = WikidataSemanticModelHelper(qnodes, wdclasses, wdprops)
        # create class nodes first
        classcount = {}
        classmap = {}
        for cid, qnode_id in cta.items():
            dnode = O.DataNode(id=f'col-{cid}', col_index=cid, label=table.table.columns[cid].name)

            # somehow, they may end-up predict multiple classes, we need to select one
            if qnode_id.find(" ") != -1:
                qnode_id = qnode_id.split(" ")[0]
            curl = sm_helper.get_qnode_uri(qnode_id)
            cnode_id = f"{curl}:{classcount.get(qnode_id, 0)}"
            classcount[qnode_id] = classcount.get(qnode_id, 0) + 1

            try:
                cnode_label = sm_helper.get_qnode_label(curl)
            except KeyError:
                cnode_label = f"wd:{qnode_id}"
            cnode = O.ClassNode(id=cnode_id, abs_uri=curl, rel_uri=f"wd:{qnode_id}", readable_label=cnode_label)
            classmap[dnode.id] = cnode.id

            sm.add_node(dnode)
            sm.add_node(cnode)
            sm.add_edge(O.Edge(source=cnode.id, target=dnode.id,
                               abs_uri=str(RDFS.label), rel_uri="rdfs:label"))

        for uid, vid, edge in sg.edges(data='data'):
            unode = sg.nodes[uid]['data']
            vnode = sg.nodes[vid]['data']

            if unode.is_column:
                suid = f"col-{unode.column}"
            else:
                suid = unode.id
            if vnode.is_column:
                svid = f"col-{vnode.column}"
            else:
                svid = vnode.id

            if sm.has_node(suid):
                if suid in classmap:
                    suid = classmap[suid]
                source = sm.get_node(suid)
            elif unode.is_column:
                # create a data node
                source = O.DataNode(id=suid, col_index=unode.column, label=table.table.get_column_by_index(unode.column).name)
                sm.add_node(source)
            else:
                # create a statement node
                source = O.ClassNode(id=suid, abs_uri='http://wikiba.se/ontology#Statement', rel_uri='wikibase:Statement')
                sm.add_node(source)
            if sm.has_node(svid):
                if svid in classmap:
                    svid = classmap[svid]
                target = sm.get_node(svid)
            elif vnode.is_column:
                target = O.DataNode(id=svid, col_index=vnode.column, label=table.table.get_column_by_index(vnode.column).name)
                sm.add_node(target)
            else:
                target = O.ClassNode(id=svid, abs_uri='http://wikiba.se/ontology#Statement',
                                     rel_uri='wikibase:Statement')
                sm.add_node(target)

            prop_uri = sm_helper.get_prop_uri(edge.predicate)
            sm.add_edge(O.Edge(source=source.id, target=target.id, abs_uri=prop_uri, rel_uri=f"p:{edge.predicate}",
                               readable_label=sm_helper.get_pnode_label(prop_uri)))

        return sm

    def get_entities(self, qnode_ids: Set[str], n_hop: int = 1, verbose: bool = False) -> Dict[str, QNode]:
        assert n_hop <= 2
        batch_size = 30
        qnodes: Dict[str, QNode] = {}
        for qnode_id in qnode_ids:
            qnode = self.qnodes.get(qnode_id, None)
            if qnode is not None:
                qnodes[qnode_id] = qnode
        
        if self.is_proxy_db:
            qnode_ids = [qnode_id for qnode_id in qnode_ids if qnode_id not in qnodes]
            if len(qnode_ids) > 0:
                resp = M.parallel_map(
                    query_wikidata_entities,
                    [qnode_ids[i:i+batch_size] for i in range(0, len(qnode_ids), batch_size)],
                    show_progress=verbose,
                    progress_desc=f'query wikidata for get entities in hop: {n_hop}',
                    is_parallel=True)
                for odict in resp:
                    for k, v in odict.items():
                        qnodes[k] = v
                        self.qnodes[k] = v

        if n_hop > 1:
            next_qnode_ids = set()
            for qnode in qnodes.values():
                for p, stmts in qnode.props.items():
                    for stmt in stmts:
                        if stmt.value.is_qnode():
                            next_qnode_ids.add(stmt.value.as_qnode_id())
                        for qvals in stmt.qualifiers.values():
                            next_qnode_ids = next_qnode_ids.union(qval.as_qnode_id() for qval in qvals if qval.is_qnode())
            next_qnode_ids = list(next_qnode_ids.difference(qnodes.keys()))
            for qnode_id in next_qnode_ids:
                qnode = self.qnodes.get(qnode_id, None)
                if qnode is not None:
                    qnodes[qnode_id] = qnode
            
            if self.is_proxy_db:
                next_qnode_ids = [qnode_id for qnode_id in next_qnode_ids if qnode_id not in qnodes]
                if len(next_qnode_ids) > 0:
                    resp = M.parallel_map(
                        query_wikidata_entities,
                        [next_qnode_ids[i:i+batch_size] for i in range(0, len(next_qnode_ids), batch_size)],
                        show_progress=verbose,
                        progress_desc=f'query wikidata for get entities in hop: {n_hop}',
                        is_parallel=True)
                    for odict in resp:
                        for k, v in odict.items():
                            qnodes[k] = v
                            self.qnodes[k] = v
        return qnodes


if __name__ == '__main__':
    cfg = OmegaConf.load(ROOT_DIR / "grams.yaml")

    # tbl = I.LinkedTable.from_csv_file(ROOT_DIR / "examples/novartis/tables/table_03.csv")
    # data = M.deserialize_json(ROOT_DIR / "examples/novartis/ground-truth/table_03/version.01.json")
    # sm = O.SemanticModel.from_dict(data['semantic_models'][0])
    # sm.draw()
    # exit(0)
    tbl = I.LinkedTable.from_dict(M.deserialize_json(ROOT_DIR / "examples/misc/tables/president_of_the_national_council_austria_10_0f1733248af445ee5a7d360a648bf9b1.json"))
    # tbl = I.W2WTable.from_csv_file(ROOT_DIR / "examples/t2dv2/tables/29414811_2_4773219892816395776.csv")
    # for ri in range(tbl.size()):
    #     for ci in range(len(tbl.table.columns)):
    #         tbl.links[ri][ci] = []
    # cea = M.deserialize_csv(ROOT_DIR / "examples/t2dv2/tables/29414811_2_4773219892816395776.candidates.tsv", delimiter="\t")
    # gold_cea = {}
    # for r in cea:
    #     ri, ci = int(r[0]), int(r[1])
    #     gold_cea[ri, ci] = r[2]
    #     tbl.links[ri][ci] = [
    #         I.Link(0, len(tbl.table.columns[ci].values[ri]), "", x.split(":")[0])
    #         for x in r[3:]
    #     ]
    grams = GRAMS(ROOT_DIR / "data", cfg)
    annotation = grams.annotate(tbl)
    annotation.sm.draw()
    print(annotation.sm.to_dict())
    grams.timer.report()

    # # %%
    # # ent column
    # from grams.algorithm.sm_wikidata import WDOnt
    # wdont = WDOnt(grams.qnodes, grams.wdclasses, grams.wdprops)
    # from collections import defaultdict
    # from grams.algorithm.data_graph import *
    # # %%
    # ci, classid = list(annotation.pred_cta.items())[0]
    # pred_ents = []
    # props = {'P136', 'P178', 'P400', 'P577'}
    # final_prediction = []
    # original_prediction = []
    # for ri in range(tbl.size()):
    #     can_ents = []
    #     for link in tbl.links[ri][ci]:
    #         if link.qnode_id is not None:
    #             qnode = grams.qnodes[link.qnode_id]
    #             if any(stmt.value.as_qnode_id() == classid for stmt in qnode.props.get("P31", [])):
    #                 can_ents.append(qnode.id)
    #
    #     qnode2score = defaultdict(set)
    #     for stmt, edges in annotation.dg[f"{ri}-{ci}"].items():
    #         assert len(edges) == 1
    #         edgeid = list(edges.keys())[0]
    #         if edgeid not in props:
    #             continue
    #         stmt = annotation.dg.nodes[stmt]['data']
    #         for target_flow in stmt.forward_flow[EdgeFlowSource(source_id=f"{ri}-{ci}", edge_id=edgeid)]:
    #             target = annotation.dg.nodes[target_flow.target_id]['data']
    #             if target.is_cell and target.column != ci:
    #                 qnode2score[stmt.qnode_id].add(target.column)
    #
    #     best_can, best_can_score = None, -1
    #     for can_ent in can_ents:
    #         can_ent_score = len(qnode2score.get(can_ent, []))
    #         if can_ent_score > best_can_score:
    #             best_can = can_ent
    #             best_can_score = can_ent_score
    #
    #     # print(ri, gold_cea[ri, ci], best_can == gold_cea[ri, ci], can_ents[0] == gold_cea[ri, ci], can_ents)
    #     print(ri, wdont.get_qnode_label(gold_cea[ri, ci]), best_can == gold_cea[ri, ci], can_ents[0] == gold_cea[ri, ci], [wdont.get_qnode_label(x) for x in can_ents])
    #     final_prediction.append((gold_cea[ri, ci], best_can))
    #     original_prediction.append((gold_cea[ri, ci], tbl.links[ri][ci][0].qnode_id))
    # #%%
    # print("TOP1", sum(x[0] == x[1] for x in final_prediction) / len(final_prediction))
    # print("TOP1 origin", sum(x[0] == x[1] for x in original_prediction) / len(final_prediction))
    #
    # count = 0
    # for ri in range(tbl.size()):
    #     if any(link.qnode_id == gold_cea[ri, ci] for link in tbl.links[ri][ci][:5]):
    #         count += 1
    # print("TOP5", count / len(final_prediction))


import itertools

import numpy as np
from enum import IntEnum
from typing import Dict, Set, List
from uuid import uuid4

from rdflib import RDFS

from grams.inputs.linked_table import LinkedTable
from grams.evaluation import sm_metrics
import sm.outputs as O
from kgdata.wikidata.models import QNode, WDProperty, WDClass
from sm.misc.graph import viz_graph
from sm.misc import identity_func


class SMNodeType(IntEnum):
    Column = 0
    Class = 1
    Statement = 2
    Entity = 3
    Literal = 4


class OutOfNamespace(Exception):
    pass


class WDOnt:
    STATEMENT_URI = "http://wikiba.se/ontology#Statement"
    STATEMENT_REL_URI = "wikibase:Statement"

    def __init__(self, qnodes: Dict[str, QNode], wdclasses: Dict[str, WDClass], wdprops: Dict[str, WDProperty]):
        self.qnodes = qnodes
        self.wdclasses = wdclasses
        self.wdprops = wdprops
        self.get_qid_fn = {
            "Q": identity_func,
            "q": identity_func,
            # http
            "h": self.get_qnode_id
        }
        self.get_pid_fn = {
            "P": identity_func,
            "p": identity_func,
            "h": self.get_prop_id
        }

    @classmethod
    def is_uri_statement(cls, uri: str):
        return uri == 'http://wikiba.se/ontology#Statement'

    @classmethod
    def is_uri_dummy_class(cls, uri: str):
        return uri == 'http://wikiba.se/ontology#DummyClassForInversion'

    @classmethod
    def is_uri_property(cls, uri: str):
        return uri.startswith(f"http://www.wikidata.org/prop/")

    @classmethod
    def is_uri_qnode(cls, uri: str):
        return uri.startswith("http://www.wikidata.org/entity/")

    @classmethod
    def get_qnode_id(cls, uri: str):
        if not cls.is_uri_qnode(uri):
            raise OutOfNamespace(f"{uri} is not in wikidata qnode namespace")
        return uri.replace("http://www.wikidata.org/entity/", "")

    @classmethod
    def get_qnode_uri(cls, qnode_id: str):
        return f"http://www.wikidata.org/entity/{qnode_id}"

    @classmethod
    def get_prop_id(cls, uri: str):
        if not cls.is_uri_property(uri):
            raise OutOfNamespace(f"{uri} is not in wikidata property namespace")
        return uri.replace(f"http://www.wikidata.org/prop/", "")

    @classmethod
    def get_prop_uri(cls, pid: str):
        return f"http://www.wikidata.org/prop/{pid}"

    def get_qnode_label(self, uri_or_id: str):
        qid = self.get_qid_fn[uri_or_id[0]](uri_or_id)
        if qid in self.wdclasses:
            return f"{self.wdclasses[qid].label} ({qid})"
        return f"{self.qnodes[qid].label} ({qid})"

    def get_pnode_label(self, uri_or_id: str):
        pid = self.get_pid_fn[uri_or_id[0]](uri_or_id)
        # TODO: fix me! should not do this
        if pid not in self.wdprops:
            return pid
        return f"{self.wdprops[pid].label} ({pid})"


class WikidataSemanticModelHelper(WDOnt):
    ID_PROPS = {str(RDFS.label)}

    def norm_sm(self, sm: O.SemanticModel):
        """ "Normalize the semantic model with the following modifications:
        1. Add readable label to edge and class
        2. Convert direct link (without statement) to have statement except the id props.
        """
        new_sm = sm.clone()

        # update readable label
        for n in new_sm.iter_nodes():
            if n.is_class_node:
                if self.is_uri_qnode(n.abs_uri):
                    n.readable_label = self.get_qnode_label(n.abs_uri)
            elif n.is_literal_node:
                if self.is_uri_qnode(n.value):
                    n.readable_label = self.get_qnode_label(n.value)
        for e in new_sm.iter_edges():
            if e.abs_uri not in self.ID_PROPS:
                e.readable_label = self.get_pnode_label(e.abs_uri)

        # convert direct link
        for edge in list(new_sm.iter_edges()):
            if edge.abs_uri in self.ID_PROPS:
                continue
            source = new_sm.get_node(edge.source)
            target = new_sm.get_node(edge.target)

            if (not source.is_class_node or (source.is_class_node and source.abs_uri != WDOnt.STATEMENT_URI)) \
                    and (not target.is_class_node or (target.is_class_node and target.abs_uri != WDOnt.STATEMENT_URI)):
                # this is direct link, we replace its edge
                assert len(new_sm.get_edges_between_nodes(source.id, target.id)) == 1
                new_sm.remove_edges_between_nodes(source.id, target.id)
                stmt = O.ClassNode(str(uuid4()), WDOnt.STATEMENT_URI, WDOnt.STATEMENT_REL_URI, False,
                                   "Statement")
                new_sm.add_node(stmt)

                new_sm.add_edge(O.Edge(source=edge.source,
                                       target=stmt.id,
                                       abs_uri=edge.abs_uri, rel_uri=edge.rel_uri,
                                       approximation=edge.approximation, readable_label=edge.readable_label))
                new_sm.add_edge(O.Edge(source=stmt.id,
                                       target=edge.target,
                                       abs_uri=edge.abs_uri, rel_uri=edge.rel_uri,
                                       approximation=edge.approximation, readable_label=edge.readable_label))
        return new_sm

    @staticmethod
    def minify_sm(sm: O.SemanticModel):
        """This is a reverse function of `norm_sm`:
        1. Remove an intermediate statement if it doesn't have any qualifiers
        """
        new_sm = sm.clone()
        for n in sm.iter_nodes():
            if n.is_class_node and n.abs_uri == WDOnt.STATEMENT_URI:
                inedges = sm.incoming_edges(n.id)
                outedges = sm.outgoing_edges(n.id)
                if len(outedges) == 1 and outedges[0].abs_uri == inedges[0].abs_uri:
                    # no qualifiers
                    new_sm.remove_node(n.id)
                    for inedge in inedges:
                        assert inedge.abs_uri == outedges[0].abs_uri
                        new_sm.add_edge(O.Edge(inedge.source, outedges[0].target, inedge.abs_uri, inedge.rel_uri,
                                               # just in case user misannotate to not include approximation in both links
                                               inedge.approximation or outedges[0].approximation,
                                               inedge.readable_label))
        return new_sm

    def gen_equivalent_sm(self, sm: O.SemanticModel, strict: bool = True, force_inversion: bool = False):
        """Given a semantic model (not being modified), generate equivalent models by inferring inverse properties.

        Parameters
        ----------
        sm: the input semantic model (original)
        strict: whether to throw exception when target of an inverse property is not a class.
        force_inversion: only work when strict mode is set to false. Without force_inverse, we skip inverse properties,
                       otherwise, we generate an inverse model with a special class: wikibase:DummyClassForInversion

        Returns
        -------

        """
        """Given an semantic model (not being modified), generate equivalent models
        by inferring inverse properties. Running on strict mode mean it will check if the invertible property is apply
        to a non-class node (column that .

        Currently, we only inverse the properties, not qualifiers.
        """
        sm = self.norm_sm(sm)
        invertible_stmts = []
        is_class_fn = lambda n1: n1.is_class_node or (n1.is_literal_node and self.is_uri_qnode(n1.value))

        for n in sm.iter_nodes():
            if n.is_class_node and WDOnt.is_uri_statement(n.abs_uri):
                inedges = sm.incoming_edges(n.id)
                outedges = sm.outgoing_edges(n.id)
                # only has one prop
                prop, = list({inedge.abs_uri for inedge in inedges})
                pid = self.get_prop_id(prop)
                stmt_has_value = False
                for outedge in outedges:
                    if outedge.abs_uri != prop:
                        # assert len(self.wdprops[self.get_prop_id(outedge.abs_uri)].inverse_properties) == 0, "Just to make sure" \
                        #                                                                                    "that qualifiers is not invertable. Otherwise, this algorithm will missing one generated SMs"
                        # character role has an inverse property: performer. They can be used as qualifier so nothing to do here just pass
                        pass
                    else:
                        stmt_has_value = True
                if len(self.wdprops[pid].inverse_properties) > 0 and stmt_has_value:
                    # invertible property
                    # people seem to misunderstand what inverse_property means in RDF;
                    # inverse doesn't apply to data property but only object property.
                    # so we catch the error here to detect what we should fix.
                    outedge, = [outedge for outedge in outedges if outedge.abs_uri == prop]
                    targets_are_class = is_class_fn(sm.get_node(outedge.target))
                    if targets_are_class:
                        invertible_stmts.append(n)
                    elif strict:
                        raise Exception(f"{pid} is not invertible")
                    elif force_inversion:
                        assert sm.get_node(outedge.target).is_data_node, "Clearly the model is wrong, you have an inverse property to a literal node"
                        invertible_stmts.append(n)

        # we have N statement, so we are going to have N! - 1 ways. It's literally a cartesian product
        all_choices = []
        for stmt in invertible_stmts:
            # assume that each statement only has one incoming link! fix the for loop if this assumption doesn't hold
            inedge, = sm.incoming_edges(stmt.id)
            choice = [(stmt, None, None)]
            for invprop in self.wdprops[self.get_prop_id(inedge.abs_uri)].inverse_properties:
                choice.append((stmt, self.get_prop_uri(invprop), f"p:{invprop}"))
            all_choices.append(choice)

        n_choices = np.prod([len(c) for c in all_choices]) - 1
        if n_choices > 256:
            raise sm_metrics.PermutationExploding("Too many possible semantic models")

        all_choices = list(itertools.product(*all_choices))
        assert all(invprop is None for _, invprop, _ in all_choices[0]), "First choice is always the current semantic model"
        new_sms = [sm]
        for choice in all_choices[1:]:
            new_sm = sm.clone()
            # we now change the statement from original prop to use the inverse prop (change direction)
            # if the invprop is not None
            for stmt, invprop_abs_uri, invprop_rel_uri in choice:
                if invprop_abs_uri is None:
                    continue
                readable_label = self.get_pnode_label(invprop_abs_uri)
                # assume that each statement only has one incoming link! fix the for loop if this assumption doesn't hold
                inedge, = sm.incoming_edges(stmt.id)
                # statement must have only one property
                outedge, = [outedge for outedge in sm.outgoing_edges(stmt.id) if outedge.abs_uri == inedge.abs_uri]
                assert len(new_sm.get_edges_between_nodes(inedge.source, stmt.id)) == 1 and \
                       len(new_sm.get_edges_between_nodes(stmt.id, outedge.target)) == 1
                new_sm.remove_edges_between_nodes(inedge.source, stmt.id)
                new_sm.remove_edges_between_nodes(stmt.id, outedge.target)

                target = sm.get_node(outedge.target)
                if not is_class_fn(target):
                    assert target.is_data_node
                    dummy_class_node = O.ClassNode(str(uuid4()), abs_uri='http://wikiba.se/ontology#DummyClassForInversion',
                                                   rel_uri='wikibase:DummyClassForInversion')
                    new_sm.add_node(dummy_class_node)
                    new_sm.add_edge(O.Edge(source=dummy_class_node.id, target=target.id, abs_uri=str(RDFS.label),
                                       rel_uri="rdfs:label"))
                    outedge_target = dummy_class_node.id
                else:
                    outedge_target = outedge.target
                new_sm.add_edge(O.Edge(source=outedge_target, target=stmt.id,
                                       abs_uri=invprop_abs_uri, rel_uri=invprop_rel_uri,
                                       approximation=outedge.approximation, readable_label=readable_label))
                new_sm.add_edge(O.Edge(source=stmt.id, target=inedge.source,
                                       abs_uri=invprop_abs_uri, rel_uri=invprop_rel_uri,
                                       approximation=inedge.approximation, readable_label=readable_label))
            new_sms.append(new_sm)
        return new_sms

    def viz_sm(self, sm: O.SemanticModel, outdir, graph_id) -> 'WikidataSemanticModelHelper':
        colors = {
            SMNodeType.Column: dict(fill='#ffd666', stroke='#874d00'),
            SMNodeType.Class: dict(fill="#b7eb8f", stroke="#135200"),
            SMNodeType.Statement: dict(fill="#d9d9d9", stroke="#434343"),
            SMNodeType.Entity: dict(fill="#C6E5FF", stroke="#5B8FF9"),
            SMNodeType.Literal: dict(fill="#C6E5FF", stroke="#5B8FF9"),
        }

        def node_fn(uid, udata):
            u: O.Node = udata['data']
            nodetype = self.get_node_type(u)
            if isinstance(u, O.DataNode):
                label = u.label
            elif isinstance(u, O.ClassNode):
                if u.abs_uri == WDOnt.STATEMENT_URI:
                    label = ""
                elif u.readable_label is None:
                    if self.is_uri_qnode(u.abs_uri):
                        label = self.get_qnode_label(u.abs_uri)
                    else:
                        label = u.label
                else:
                    label = u.readable_label
            else:
                label = u.readable_label

            return {
                "label": label,
                "style": colors[nodetype],
                "labelCfg": {
                    "style": {
                        "fill": "black",
                        "background": {
                            "padding": [4, 4, 4, 4],
                            "radius": 3,
                            **colors[nodetype]
                        }
                    }
                },
            }

        def edge_fn(eid, edata):
            edge: O.Edge = edata['data']
            label = edge.readable_label
            if label is None:
                if edge.abs_uri not in self.ID_PROPS:
                    label = self.get_pnode_label(edge.abs_uri)
                else:
                    label = edge.label

            return {
                "label": label
            }

        viz_graph(sm.g, node_fn, edge_fn, outdir, graph_id)
        return self

    def get_node_type(self, n: O.Node):
        if isinstance(n, O.LiteralNode):
            if self.is_uri_qnode(n.value):
                return SMNodeType.Entity
            return SMNodeType.Literal
        if isinstance(n, O.DataNode) or (isinstance(n, O.ClassNode) and self.is_uri_column(n.abs_uri)):
            return SMNodeType.Column
        if isinstance(n, O.ClassNode):
            if n.abs_uri == WDOnt.STATEMENT_URI:
                return SMNodeType.Statement
            return SMNodeType.Class
        raise Exception("Unreachable!")

    def get_entity_columns(self, sm: O.SemanticModel) -> List[int]:
        ent_columns = []
        for dnode in sm.iter_nodes():
            if dnode.is_data_node:
                inedges = sm.incoming_edges(dnode.id)
                if len(inedges) == 0:
                    continue
                assert len({edge.abs_uri for edge in inedges}) == 1, inedges
                edge_abs_uri = inedges[0].abs_uri
                if edge_abs_uri in self.ID_PROPS:
                    assert len(inedges) == 1, inedges
                    source = sm.get_node(inedges[0].source)
                    assert not self.is_uri_statement(source.abs_uri)
                    ent_columns.append(dnode.col_index)
        return ent_columns

    @classmethod
    def is_uri_column(cls, uri: str):
        """Test if an uri is for specifying the column"""
        return uri.startswith("http://example.com/table/")

    @staticmethod
    def get_column_uri(column_index: int):
        return f"http://example.com/table/{column_index}"

    @staticmethod
    def get_column_index(uri: str):
        assert WikidataSemanticModelHelper.is_uri_column(uri)
        return int(uri.replace("http://example.com/table/", ""))

    def extract_claims(self, tbl: LinkedTable, sm: O.SemanticModel, allow_multiple_ent: bool = True):
        """Extract claims from the table given a semantic model.

        If an entity doesn't have link, its id will be null
        """
        # norm the semantic model first
        sm = self.norm_sm(sm)
        schemas = {}
        for u in sm.iter_nodes():
            if not u.is_class_node or WDOnt.is_uri_statement(u.abs_uri):
                continue

            schema = {"props": {}, "subject": None, "sm_node_id": u.id}
            for us_edge in sm.outgoing_edges(u.id):
                if us_edge.abs_uri in self.ID_PROPS:
                    v = sm.get_node(us_edge.target)
                    assert v.is_data_node
                    assert schema['subject'] is None
                    schema['subject'] = v.col_index
                    continue

                s = sm.get_node(us_edge.target)
                assert s.is_class_node and WDOnt.is_uri_statement(s.abs_uri)
                assert WDOnt.is_uri_property(us_edge.abs_uri)

                pnode = WDOnt.get_prop_id(us_edge.abs_uri)
                if pnode not in schema['props']:
                    schema['props'][pnode] = []

                stmt = {
                    "index": len(schema['props'][pnode]),
                    'value': None,
                    'qualifiers': []
                }
                schema['props'][pnode].append(stmt)
                for sv_edge in sm.outgoing_edges(s.id):
                    v = sm.get_node(sv_edge.target)

                    assert WDOnt.is_uri_property(sv_edge.abs_uri)
                    if sv_edge.abs_uri == us_edge.abs_uri:
                        assert stmt['value'] is None, "only one property"
                        # this is property
                        if v.is_class_node:
                            stmt['value'] = {'type': 'classnode', 'value': v.id}
                        elif v.is_data_node:
                            stmt['value'] = {'type': 'datanode', 'value': v.col_index}
                        else:
                            assert v.is_literal_node
                            stmt['value'] = {'type': 'literalnode', 'value': v.value}
                    else:
                        # this is qualifier
                        if v.is_class_node:
                            stmt['qualifiers'].append({'type': 'classnode', 'pnode': WDOnt.get_prop_id(sv_edge.abs_uri), 'value': v.id})
                        elif v.is_data_node:
                            stmt['qualifiers'].append({'type': 'datanode', 'pnode': WDOnt.get_prop_id(sv_edge.abs_uri), 'value': v.col_index})
                        else:
                            assert v.is_literal_node
                            stmt['qualifiers'].append({'type': 'literalnode', 'pnode': WDOnt.get_prop_id(sv_edge.abs_uri), 'value': v.value})
            schemas[u.id] = schema

        assert all(c.index == ci for ci, c in enumerate(tbl.table.columns)), "Cannot handle table with missing columns yet"

        records = [{} for ri in range(len(tbl.table.columns[0].values))]
        node2ents = {}

        # extract data props first
        for cid, schema in schemas.items():
            ci = schema['subject']
            col = tbl.table.columns[ci]
            for ri, val in enumerate(col.values):
                # get entities
                qnode_ids = sorted({e.qnode_id for e in tbl.links[ri][ci] if e.qnode_id is not None and e.start < e.end})
                if len(qnode_ids) == 0:
                    # create new entity
                    ents = [{
                        "id": f"{ri}-{ci}",
                        "column": ci,
                        "row": ri,
                        "uri": None,
                        "label": val,
                        "props": {}
                    }]
                else:
                    ents = [
                        {
                            "id": qnode_id, "uri": WDOnt.get_qnode_uri(qnode_id),
                            "label": self.qnodes[qnode_id].label, "props": {}
                        }
                        for qnode_id in qnode_ids
                    ]

                if len(ents) > 1:
                    if not allow_multiple_ent:
                        raise Exception("Encounter multiple entities")

                for prop, stmts in schema['props'].items():
                    for ent in ents:
                        assert prop not in ent['props']
                        ent['props'][prop] = [{'value': None, 'qualifiers': {}} for stmt in stmts]
                        for stmt in stmts:
                            # set statement property
                            if stmt['value']['type'] == 'classnode':
                                # do it in later phase
                                pass
                            elif stmt['value']['type'] == 'datanode':
                                tci = stmt['value']['value']
                                ent['props'][prop][stmt['index']]['value'] = tbl.table.columns[tci].values[ri]
                            else:
                                assert stmt['value']['type'] == 'literalnode'
                                ent['props'][prop][stmt['index']]['value'] = stmt['value']['value']

                            # set statement qualifiers
                            for qual in stmt['qualifiers']:
                                if qual['pnode'] not in ent['props'][prop][stmt['index']]['qualifiers']:
                                    ent['props'][prop][stmt['index']]['qualifiers'][qual['pnode']] = []
                                if qual['type'] == 'classnode':
                                    # do it in later phase
                                    pass
                                elif qual['type'] == 'datanode':
                                    tci = qual['value']
                                    ent['props'][prop][stmt['index']]['qualifiers'][qual['pnode']].append(tbl.table.columns[tci].values[ri])
                                elif qual['type'] == 'literalnode':
                                    ent['props'][prop][stmt['index']]['qualifiers'][qual['pnode']].append(qual['value'])

                for ent in ents:
                    assert (ent['id'], ci) not in records[ri]
                    records[ri][ent['id'], ci] = ent
                node2ents[schema['sm_node_id'], ri] = [ent for ent in ents]

        for cid, schema in schemas.items():
            ci = schema['subject']
            col = tbl.table.columns[ci]
            for ri in range(len(col.values)):
                ulst = node2ents[schema['sm_node_id'], ri]
                for prop, stmts in schema['props'].items():
                    for stmt in stmts:
                        if stmt['value']['type'] == 'classnode':
                            vlst = node2ents[stmt['value']['value'], ri]
                            for u in ulst:
                                assert len(vlst) > 0
                                u['props'][prop][stmt['index']]['value'] = vlst[0]['id']
                                if len(vlst) > 1:
                                    # this statement must not have other qualifiers, so that v can be a list
                                    # and we can create extra statement
                                    assert len(stmt['qualifiers']) == 0
                                    for v in vlst[1:]:
                                        u['props'][prop].append({'value': v['id'], 'qualifiers': {}})
                        for qual in stmt['qualifiers']:
                            if qual['type'] == 'classnode':
                                for u in ulst:
                                    vlst = node2ents[qual['value'], ri]
                                    u['props'][prop][stmt['index']]['qualifiers'][qual['pnode']] = [v['id'] for v in vlst]

        new_records = []
        for ri, record in enumerate(records):
            new_record = {}
            for (ent_id, ci), ent in record.items():
                if ent_id in new_record:
                    # merge the entity
                    for pid, _stmts in ent['props'].items():
                        if pid not in new_record[ent_id]['props']:
                            new_record[ent_id]['props'][pid] = _stmts
                        else:
                            for _stmt in _stmts:
                                if not any(x == _stmt for x in new_record[ent_id]['props'][pid]):
                                    new_record[ent_id]['props'][pid].append(_stmt)
                else:
                    new_record[ent_id] = ent
            new_records.append(new_record)
        return records






from typing import List, Dict, Tuple, Callable, Any, Optional

import grams.misc as M
from grams.config import DATA_DIR
from grams.kg_data.wikidatamodels import WDProperty, QNode
from grams.main import GRAMS


def update_props(props: List[str]):
    props: Dict[str, QNode] = GRAMS._query_wikidata_entities(props)
    ser = []
    for p in props.values():
        np = WDProperty(p.id, str(p.label), str(p.description), str(p.datatype), [str(s) for s in p.aliases],
                   sorted({stmt.value.as_qnode_id() for stmt in p.props.get("P1647", [])}),
                   sorted({stmt.value.as_qnode_id() for stmt in p.props.get("P1659", [])}),
                   sorted({stmt.value.as_string() for stmt in p.props.get("P1628", [])}),
                   sorted({stmt.value.as_qnode_id() for stmt in p.props.get("P1629", [])}),
                   sorted({stmt.value.as_qnode_id() for stmt in p.props.get("P1696", [])}),
                   sorted({stmt.value.as_qnode_id() for stmt in p.props.get("P31", [])}))
        ser.append(np.serialize())
    M.serialize_byte_lines(ser, DATA_DIR / "new_properties.jl")


if __name__ == '__main__':
    update_props(["P8901"])



from collections import defaultdict
from typing import Dict, List

import networkx as nx
import numpy as np

from grams.algorithm.data_graph import CellNode
from grams.algorithm.literal_match import TextParser
from grams.algorithm.semantic_graph import SGColumnNode
from grams.inputs.linked_table import LinkedTable
from kgdata.wikidata.models import QNode, WDProperty, WDQuantityPropertyStats


class TypeFeatureExtraction:
    Freq = "FrequencyOfType"
    FreqOverRow = "FreqOfTypeOverRow"

    def __init__(self,
                 table: LinkedTable, sg: nx.MultiDiGraph, dg: nx.MultiDiGraph,
                 qnodes: Dict[str, QNode], wdprops: Dict[str, WDProperty],
                 wd_num_prop_stats: Dict[str, WDQuantityPropertyStats]):
        self.table = table
        self.sg = sg
        self.dg = dg
        self.qnodes = qnodes
        self.wdprops = wdprops
        self.wd_num_prop_stats = wd_num_prop_stats
        self.text_parser = TextParser()

        # self.transitive_props = [p.id for p in self.wdprops.values() if p.is_transitive()]
        self.hierarchy_props = {"P131", "P276"}
    
    def extract_features(self):
        freq_type = {}
        freq_over_row = {}
        cell2qnodes = {}
        column2types = {}

        for uid, udata in self.sg.nodes(data=True):
            u: SGColumnNode = udata['data']
            if not u.is_column:
                continue

            # cells in this column
            cells: List[CellNode] = [self.dg.nodes[cid]['data'] for cid in u.nodes]
            covered_fractions = [
                sum(span.length for spans in cell.qnodes_span.values() for span in spans) / max(len(cell.value), 1)
                for cell in cells
                if len(cell.qnode_ids) > 0
            ]
            if len(covered_fractions) == 0:
                continue
            avg_covered_fractions = np.mean(covered_fractions)
            if avg_covered_fractions < 0.8:
                continue

            for cell in cells:
                self.add_merge_qnodes(cell, cell2qnodes)

            type2freq = defaultdict(int)
            for cell in cells:
                classes = set()
                for qnode in cell2qnodes[cell.id]:
                    for stmt in qnode.props.get("P31", []):
                        classes.add(stmt.value.as_qnode_id())
                for c in classes:
                    type2freq[c] += 1

            for c, freq in type2freq.items():
                freq_type[uid, c] = freq
                freq_over_row[uid, c] = freq / self.table.size()
            column2types[uid] = list(type2freq.keys())

        return {
            self.Freq: freq_type,
            self.FreqOverRow: freq_over_row,
            "_column_to_types": column2types
        }
    
    def add_merge_qnodes(self, cell: CellNode, cell2qnodes: Dict[str, List[QNode]]):
        # merge qnodes that are sub of each other
        # attempt to merge qnodes (spatial) if they are contained in each other
        # we should go even higher order
        if len(cell.qnode_ids) > 1:
            # attempt to merge qnodes (spatial) if they are contained in each other
            # we should go even higher order
            ignore_qnodes = set()
            for q0_id in cell.qnode_ids:
                q0 = self.qnodes[q0_id]
                vals = {
                    stmt.value.as_qnode_id()
                    for p in self.hierarchy_props
                    for stmt in q0.props.get(p, [])
                }
                for q1_id in cell.qnode_ids:
                    if q0_id == q1_id:
                        continue
                    if q1_id in vals:
                        # q0 is inside q1, ignore q1
                        ignore_qnodes.add(q1_id)
            cell2qnodes[cell.id] = [self.qnodes[q_id] for q_id in cell.qnode_ids if q_id not in ignore_qnodes]
        elif len(cell.qnode_ids) > 0:
            cell2qnodes[cell.id] = [self.qnodes[cell.qnode_ids[0]]]
        else:
            cell2qnodes[cell.id] = []


from dataclasses import dataclass, asdict
from hashlib import md5
from pathlib import Path
from typing import List, Optional, Union
from urllib.parse import urlparse

import orjson
from slugify import slugify

import sm.misc as M
from sm.inputs.table import ColumnBasedTable, Column


@dataclass
class LinkedTable:
    # the table that we are working on
    table: ColumnBasedTable

    context: "Context"

    # a mapping from (row id, column id) to the list of links attach to that cell
    links: List[List[List["Link"]]]

    @property
    def id(self):
        return self.table.table_id

    def size(self):
        if len(self.table.columns) == 0:
            return 0
        return len(self.table.columns[0].values)

    def to_json(self):
        """This function convert the current table into a dictionary to store in json
        Note that this function doesn't store the features since it's not all kinds of features are json serializable
        """
        return {
            "table": self.table.to_dict(),
            "context": asdict(self.context),
            "links": [
                [[asdict(link) for link in links] for links in rlinks]
                for rlinks in self.links
            ],
        }

    def get_friendly_fs_id(self):
        id = self.id
        if id.startswith("http://") or id.startswith("https://"):
            if id.find("dbpedia.org") != -1:
                id = slugify(
                    urlparse(id).path.replace("/resource/", "").replace("/", "_")
                ).replace("-", "_")
                id += "_" + md5(self.id.encode()).hexdigest()
            elif id.find("wikipedia.org") != -1:
                id = slugify(
                    urlparse(id).path.replace("/wiki/", "").replace("/", "_")
                ).replace("-", "_")
                id += "_" + md5(self.id.encode()).hexdigest()
            else:
                raise NotImplementedError()

        return id

    @staticmethod
    def from_dict(odict: dict):
        tbl = ColumnBasedTable.from_dict(odict["table"])
        context = Context(**odict["context"])
        links = [
            [[Link(**link) for link in links] for links in rlinks]
            for rlinks in odict["links"]
        ]
        return LinkedTable(tbl, context, links)

    @staticmethod
    def from_column_based_table(tbl: ColumnBasedTable):
        links = []
        if len(tbl.columns) > 0:
            links = [
                [[] for ci in range(len(tbl.columns))]
                for ri in range(len(tbl.columns[0].values))
            ]
        return LinkedTable(tbl, Context(None, None, None), links)

    @staticmethod
    def from_csv_file(
        infile: Union[Path, str],
        link_file: Optional[str] = None,
        first_row_header: bool = True,
        table_id: Optional[str] = None,
    ):
        infile = Path(infile)
        if link_file is None:
            link_file = infile.parent / f"{infile.stem}.links.tsv"
        else:
            link_file = Path(link_file)

        if table_id is None:
            table_id = infile.stem
        rows = M.deserialize_csv(infile)

        assert len(rows) > 0, "Empty table"
        columns = []
        if first_row_header:
            headers = rows[0]
            rows = rows[1:]
        else:
            headers = [f"column-{i:03d}" for i in range(len(rows[0]))]

        for ci, cname in enumerate(headers):
            columns.append(Column(ci, cname, [r[ci] for r in rows]))
        table = ColumnBasedTable(table_id, columns)
        links = []
        for ri in range(len(rows)):
            links.append([[] for _ci in range(len(headers))])

        if link_file.exists():
            for row in M.deserialize_csv(link_file, delimiter="\t"):
                ri, ci, ents = int(row[0]), int(row[1]), row[2:]
                for ent in ents:
                    if ent.startswith("{"):
                        # it's json, encoding the hyperlinks
                        link = Link(**orjson.loads(ent))
                    elif ent.startswith("http"):
                        assert ent.startswith("http://www.wikidata.org/entity/")
                        link = Link(
                            0,
                            len(table.columns[ci][ri]),
                            ent,
                            ent.replace("http://www.wikidata.org/entity/", ""),
                        )
                    else:
                        link = Link(
                            0,
                            len(table.columns[ci][ri]),
                            f"http://www.wikidata.org/entity/{ent}",
                            ent,
                        )
                    links[ri][ci].append(link)
        return LinkedTable(table, Context(), links)


@dataclass
class Context:
    page_title: Optional[str] = None
    page_url: Optional[str] = None
    page_qnode: Optional[str] = None


@dataclass
class Link:
    start: int
    end: int
    url: str
    qnode_id: Optional[str]



import copy
import numpy as np
from collections import Counter
from dataclasses import dataclass
from itertools import chain, combinations
from operator import itemgetter
from typing import Dict, Optional, Union, List, Set, Tuple, NamedTuple, Any, Callable
from uuid import uuid4

import networkx as nx
from rdflib import RDFS
from typing_extensions import TypedDict
from grams.algorithm.sm_wikidata import WDOnt
from kgdata.wikidata.models import QNode, DataValue, WDProperty, WDClass
from sm.misc.graph import viz_graph
from grams.inputs.linked_table import LinkedTable
from grams.algorithm.data_graph import build_data_graph, DGNode, StatementNode, CellNode, \
    ContextSpan, EdgeFlowSource, EdgeFlowTarget, \
    FlowProvenance
import sm.outputs as O


@dataclass
class SGColumnNode:
    id: str
    # column name
    label: str
    # column index
    column: int
    # list nodes' id in the data graph
    nodes: Set[str]

    @property
    def is_column(self):
        return True

    @property
    def is_value(self):
        return False

    @property
    def is_statement(self):
        return False

    def clone(self):
        return copy.deepcopy(self)


@dataclass
class SGEntityValueNode:
    id: str
    label: str
    qnode_id: str
    context_span: Optional[ContextSpan]

    @property
    def is_column(self):
        return False

    @property
    def is_value(self):
        return True

    @property
    def is_entity_value(self):
        return True

    @property
    def is_literal_value(self):
        return False

    @property
    def is_in_context(self):
        return self.context_span is not None

    @property
    def is_statement(self):
        return False

    def clone(self):
        return copy.deepcopy(self)


@dataclass
class SGLiteralValueNode:
    id: str
    label: str
    value: DataValue
    context_span: Optional[ContextSpan]

    @property
    def is_column(self):
        return False

    @property
    def is_value(self):
        return True

    @property
    def is_entity_value(self):
        return False

    @property
    def is_literal_value(self):
        return True

    @property
    def is_in_context(self):
        return self.context_span is not None

    @property
    def is_statement(self):
        return False

    def clone(self):
        return copy.deepcopy(self)


SGEdgeFlowTarget = NamedTuple("SGEdgeFlowTarget", [("dg_target_id", str), ("sg_target_id", str), ("edge_id", str)])
# the source of the flow is not needed but still there for what reason? consistent?
SGEdgeFlowSource = NamedTuple("SGEdgeFlowSource", [("dg_source_id", str), ("sg_source_id", str), ("edge_id", str)])


@dataclass
class SGStatementNode:
    id: str
    # (source flow, target_flow) => dg statement id => provenance (we have multiple statement id because
    # for one qnode one property, we may have multiple matches statement
    forward_flow: Dict[SGEdgeFlowSource, Dict[SGEdgeFlowTarget, Set[str]]]
    reversed_flow: Dict[SGEdgeFlowTarget, Dict[SGEdgeFlowSource, Set[str]]]
    flow: Dict[Tuple[SGEdgeFlowSource, SGEdgeFlowTarget], Dict[str, List[FlowProvenance]]]

    @staticmethod
    def get_id(source_id: str, source_predicate: str, target_id: str):
        return f"stmt:{source_id}-{source_predicate}-{target_id}"

    @staticmethod
    def new(id: str):
        return SGStatementNode(id, {}, {}, {})

    @property
    def is_column(self):
        return False

    @property
    def is_value(self):
        return False

    @property
    def is_statement(self):
        return True

    def clone(self):
        return copy.deepcopy(self)

    def track_provenance(self, source_flow: SGEdgeFlowSource, sid: str, target_flow: SGEdgeFlowTarget, provenances: List[FlowProvenance]):
        """Track the provenance of which edges goes through the node"""
        if source_flow not in self.forward_flow:
            self.forward_flow[source_flow] = {}
        if target_flow not in self.forward_flow[source_flow]:
            self.forward_flow[source_flow][target_flow] = set()
        self.forward_flow[source_flow][target_flow].add(sid)

        if target_flow not in self.reversed_flow:
            self.reversed_flow[target_flow] = {}
        if source_flow not in self.reversed_flow[target_flow]:
            self.reversed_flow[target_flow][source_flow] = set()
        self.reversed_flow[target_flow][source_flow].add(sid)

        if (source_flow, target_flow) not in self.flow:
            self.flow[source_flow, target_flow] = {}
        assert sid not in self.flow[source_flow, target_flow], "Gate to make sure the assumption (source_flow, target_flow, dg statement id) is unique correct"
        self.flow[source_flow, target_flow][sid] = provenances

    def compute_freq(self, _sg: nx.MultiDiGraph, dg: nx.MultiDiGraph, edge: 'SGEdge', is_unique_freq: bool):
        """Compute the frequency of data edges between """
        if not is_unique_freq:
            freq = 0
            if edge.target_id == self.id:
                # from node to statement, we only consider the target that is property of the statement
                # since we always create new statement based on the value
                for (source_flow, target_flow), stmt2prov in self.flow.items():
                    if source_flow.edge_id != edge.predicate or source_flow.sg_source_id != edge.source_id:
                        continue
                    if target_flow.edge_id != edge.predicate:
                        continue
                    # same_stmt = False
                    # for stmt_id, prov in stmt2prov.items():
                    #     stmt: StatementNode = dg.nodes[stmt_id]['data']
                    #     if (EdgeFlowSource(source_flow.dg_source_id, source_flow.edge_id),
                    #         EdgeFlowTarget(target_flow.dg_target_id, target_flow.edge_id)) in stmt.flow:
                    #         same_stmt = True
                    #         break
                    #
                    # if same_stmt:
                    freq += 1
            else:
                for (source_flow, target_flow), stmt2prov in self.flow.items():
                    if target_flow.sg_target_id == edge.target_id and target_flow.edge_id == edge.predicate:
                        freq += 1
            return freq

        unique_pairs = set()
        if edge.target_id == self.id:
            for (source_flow, target_flow), stmt2prov in self.flow.items():
                if source_flow.sg_source_id != edge.source_id or source_flow.edge_id != edge.predicate:
                    continue

                # the target is the one contain value of the statement property, which we should have
                # only one
                if target_flow.edge_id != edge.predicate:
                    continue

                dg_source = dg.nodes[source_flow.dg_source_id]['data']
                if dg_source.is_cell:
                    source_val = dg_source.value
                else:
                    assert dg_source.is_entity_value
                    source_val = dg_source.qnode_id

                dg_target = dg.nodes[target_flow.dg_target_id]['data']
                if dg_target.is_cell:
                    target_val = dg_target.value
                elif dg_target.is_literal_value:
                    target_val = dg_target.value.to_string_repr()
                else:
                    assert dg_target.is_entity_value
                    target_val = dg_target.qnode_id

                unique_pairs.add((source_val, target_val))
        else:
            for source_flow, target_flows in self.forward_flow.items():
                dg_source = dg.nodes[source_flow.dg_source_id]['data']
                if dg_source.is_cell:
                    source_val = dg_source.value
                else:
                    assert dg_source.is_entity_value
                    source_val = dg_source.qnode_id
                for target_flow in target_flows.keys():
                    if target_flow.sg_target_id == edge.target_id and target_flow.edge_id == edge.predicate:
                        dg_target = dg.nodes[target_flow.dg_target_id]['data']
                        if dg_target.is_cell:
                            target_val = dg_target.value
                        elif dg_target.is_literal_value:
                            target_val = dg_target.value.to_string_repr()
                        else:
                            assert dg_target.is_entity_value
                            target_val = dg_target.qnode_id
                        unique_pairs.add((source_val, target_val))

        return len(unique_pairs)

    def get_edges_provenance(self, edges: List['SGEdge']):
        """Retrieve all flow provenances that connects these edges.

        Note that in case we have more than one statement per flow, we merge the statements (i.e., merge
        the provenances) to return the best statements (since the provenance store how the link is generated,
        merged the provenance will select the best one).
        """
        select_target_flows = {(e.target_id, e.predicate) for e in edges}
        links = {}
        for source_flow, target_flows in self.forward_flow.items():
            # source_flow.predicate is always the same, just change to different data node
            # get list of statements that contains all edges
            stmts: Dict[str, Dict[SGEdgeFlowTarget, List[FlowProvenance]]] = None
            for target_flow in target_flows:
                if (target_flow.sg_target_id, target_flow.edge_id) not in select_target_flows:
                    continue
                if stmts is None:
                    stmts = {}
                    for stmt_id, provs in self.flow[source_flow, target_flow].items():
                        stmts[stmt_id] = {}
                        stmts[stmt_id][target_flow] = provs
                else:
                    # do intersection because we want to find the stmt contain all edges
                    stmt2provenances = self.flow[source_flow, target_flow]
                    remove_stmt_ids = set(stmts.keys()).difference(stmt2provenances.keys())
                    for stmt_id in remove_stmt_ids:
                        stmts.pop(stmt_id)
                    for stmt_id, provenances in stmt2provenances.items():
                        if stmt_id not in stmts:
                            continue
                        # no merge since the target flow is unique
                        stmts[stmt_id][target_flow] = provenances

            if stmts is not None and len(stmts) > 0:
                # do a merge to get the provenances (this is the strategy to return the best statement)
                # note that all target flows will be appear, so the first step is to reverse target flow with statement id
                swap_stmts: Dict[SGEdgeFlowTarget, List[FlowProvenance]] = {
                    target_flow: []
                    for target_flow in next(iter(stmts.values()))
                }
                for stmt_id, _target_flows in stmts.items():
                    for target_flow, provenances in _target_flows.items():
                        swap_stmts[target_flow] = FlowProvenance.merge_lst(swap_stmts[target_flow], provenances)
                links[source_flow] = swap_stmts
        return links


@dataclass
class SGEdge:
    source_id: str
    target_id: str
    predicate: str
    features: Dict[str, Any]

    def clone(self):
        return copy.deepcopy(self)


SGNode = Union[SGColumnNode, SGEntityValueNode, SGLiteralValueNode, SGStatementNode]
NxSGNodeAttr = TypedDict('NxSGNodeAttr', data=SGNode)
NxSGEdgeAttr = TypedDict('NxSGEdgeAttr', data=SGEdge)
NxSGEdge = Tuple[str, str, str, NxSGEdgeAttr]


def get_label(porq: str, qnodes: Dict[str, QNode], wdclasses: Dict[str, WDClass], wdprops: Dict[str, WDProperty]):
    if porq.startswith("Q"):
        if porq in wdclasses:
            item = wdclasses[porq]
        elif porq in qnodes:
            item = qnodes[porq]
    else:
        item = wdprops[porq]

    return f"{item.label} ({item.id})"


def merge_connected_components(graphs: List[nx.MultiDiGraph]):
    g = nx.MultiDiGraph()
    for gi in graphs:
        for uid, udata in gi.nodes.items():
            assert uid not in g.nodes
            g.add_node(uid, **udata)

        for uid, vid, eid, edge_data in gi.edges(data=True, keys=True):
            g.add_edge(uid, vid, key=eid, **edge_data)
    return g


@dataclass
class SemanticGraphConstructorArgs:
    table: LinkedTable
    dg: nx.MultiDiGraph
    sg: nx.MultiDiGraph
    # column type assignment, from column index (must stored as string) => QNode
    cta: Optional[Dict[str, str]] = None
    # semantic model
    sm: Optional[O.SemanticModel] = None


# noinspection PyMethodMayBeStatic
class SemanticGraphConstructor:
    STATEMENT_URI = WDOnt.STATEMENT_URI
    STATEMENT_REL_URI = WDOnt.STATEMENT_REL_URI

    def __init__(self, steps: List, qnodes: Dict[str, QNode], wdclasses: Dict[str, WDClass], wdprops: Dict[str, WDProperty]):
        self.steps = steps
        self.qnodes = qnodes
        self.wdclasses = wdclasses
        self.wdprops = wdprops

    def run(self, table: LinkedTable, dg: nx.MultiDiGraph, debug=False):
        args = SemanticGraphConstructorArgs(table, dg, None)
        for i, step in enumerate(self.steps):
            step(self, args)
            if debug:
                viz_sg(args.sg, self.qnodes, self.wdclasses, self.wdprops, HOME_DIR / "graph_viz" / "debug", f"g_{i:02}")
        return args

    @staticmethod
    def get_sg_node_id(u: DGNode):
        if u.is_cell:
            return f"column-{u.column}"
        return u.id

    def get_sg_node_label(self, args: SemanticGraphConstructorArgs, u: DGNode):
        if u.is_cell:
            return args.table.table.columns[u.column].name
        if u.is_entity_value:
            qnode = self.qnodes[u.qnode_id]
            return f"{qnode.label} ({qnode.id})"
        if u.is_literal_value:
            return u.value.to_string_repr()
        raise M.UnreachableError("This function doesn't supposed to be called with statement node")

    @staticmethod
    def get_sg_subgraph(sg: nx.MultiDiGraph, edges: List[Tuple[str, str, str]]):
        """Get a subgraph of the semantic graph containing a set of edges"""
        sg_prime = nx.MultiDiGraph()
        for uid, vid, eid in edges:
            if uid not in sg_prime.nodes:
                sg_prime.add_node(uid, data=sg.nodes[uid]['data'])
            if vid not in sg_prime.nodes:
                sg_prime.add_node(vid, data=sg.nodes[vid]['data'])
            sg_prime.add_edge(uid, vid, key=eid, data=sg.edges[uid, vid, eid]['data'])
        return sg_prime
    
    @staticmethod
    def keep_one_simple_path_between_important_nodes(sg: nx.MultiDiGraph,
                                                     ranking_fn: Optional[Callable[[List[List[Tuple[str, str, str]]]], List[Tuple[str, str, str]]]] = None,
                                                     both_direction: bool = True):
        """
        Let important nodes be columns and context values. If there is more than one path between the important nodes,
        we only keep one. Note that this function doesn't remove a literal value if it is value of a property of a statement
        that appear in the chosen path.

        The default ranking function is select the path with shorter length. Supply your own function for better result.
        The both direction function will select the correct one
        """
        # figure out the important nodes
        important_nodes = set()
        for uid, udata in sg.nodes(data=True):
            u: SGNode = udata['data']
            if u.is_column or (u.is_value and u.is_in_context):
                important_nodes.add(uid)

        if ranking_fn is None:
            ranking_fn = lambda paths: sorted(paths, key=lambda path: len(path))[0]

        # select one path if there are multiple edges
        selected_edges = set()
        for uid, vid in combinations(important_nodes, 2):
            forward_paths = [
                path
                for path in nx.all_simple_edge_paths(sg, uid, vid, cutoff=6)
                if not any(edge[0] in important_nodes for edge in path[1:])
            ]
            backward_paths = [
                path
                for path in nx.all_simple_edge_paths(sg, vid, uid, cutoff=6)
                if not any(edge[0] in important_nodes for edge in path[1:])
            ]
            if both_direction:
                if len(forward_paths) + len(backward_paths) > 0:
                    selected_edges = selected_edges.union(ranking_fn(forward_paths + backward_paths))
            else:
                if len(forward_paths) > 0:
                    selected_edges = selected_edges.union(ranking_fn(forward_paths))
                if len(backward_paths) > 0:
                    selected_edges = selected_edges.union(ranking_fn(backward_paths))

        new_sg = SemanticGraphConstructor.get_sg_subgraph(sg, selected_edges)

        # add missing statement values
        for sid, sdata in list(new_sg.nodes(data=True)):
            stmt: SGStatementNode = sdata['data']
            if not stmt.is_statement:
                continue

            (uid, _, eid), = list(new_sg.in_edges(sid, keys=True))
            for vid, us_edges in new_sg[sid].items():
                if eid in us_edges:
                    # we have the statement property in the graph
                    break
            else:
                # we do not have the statement property in the graph. get it from the original graph
                for vid, us_edges in sg[sid].items():
                    if eid in us_edges:
                        if vid not in new_sg.nodes:
                            new_sg.add_node(vid, data=sg.nodes[vid]['data'])
                        new_sg.add_edge(sid, vid, key=eid, data=us_edges[eid]['data'])

        return new_sg

    @staticmethod
    def st_terminal_nodes(sg: nx.MultiDiGraph):
        terminal_nodes = set()
        for uid, udata in sg.nodes(data=True):
            u: SGNode = udata['data']
            if u.is_column:
                terminal_nodes.add(uid)
        return terminal_nodes

    # ##################################################################################################################
    # CODE FOR PIPELINE
    # ##################################################################################################################
    def init_sg(self, args: SemanticGraphConstructorArgs):
        sg = nx.MultiDiGraph()
        dg = args.dg

        # first step: add node to the graph
        for uid, udata in dg.nodes(data=True):
            u: DGNode = udata['data']
            if u.is_statement:
                # we can have more than one predicates per entity column, these are represented in the statement
                continue
            
            sgi = self.get_sg_node_id(u)
            if not sg.has_node(sgi):
                label = self.get_sg_node_label(args, u)
                if u.is_cell:
                    sgu = SGColumnNode(id=sgi, label=label, column=u.column, nodes=set())
                elif u.is_entity_value:
                    sgu = SGEntityValueNode(id=sgi, label=label, qnode_id=u.qnode_id, context_span=u.context_span)
                else:
                    assert u.is_literal_value
                    sgu = SGLiteralValueNode(id=sgi, label=label, value=u.value, context_span=u.context_span)
                sg.add_node(sgi, data=sgu)

            if sg.nodes[sgi]['data'].is_column:
                sg.nodes[sgi]['data'].nodes.add(uid)

        # second step: add link
        for uid, udata in dg.nodes(data=True):
            u: DGNode = udata['data']
            if u.is_statement:
                continue

            # add statement
            p2stmts: Dict[str, List[StatementNode]] = {}
            for _, sid, us_eid, us_edata in dg.out_edges(uid, data=True, keys=True):
                if us_eid not in p2stmts:
                    p2stmts[us_eid] = []
                p2stmts[us_eid].append(dg.nodes[sid]['data'])

            sgi = self.get_sg_node_id(u)
            for p, stmts in p2stmts.items():
                for stmt in stmts:
                    # we duplicate based on SG column node, not by DG node
                    children_values = {}
                    children_non_values = []

                    for _, vid, sv_eid, sv_edata in dg.out_edges(stmt.id, data=True, keys=True):
                        v: DGNode = dg.nodes[vid]['data']
                        if (vid, sv_eid) not in stmt.forward_flow[uid, p]:
                            # because of re-use statement, we need to only consider the children of this u node only
                            continue

                        if sv_eid == p:
                            tgi = self.get_sg_node_id(v)
                            if tgi not in children_values:
                                children_values[tgi] = []
                            children_values[tgi].append(v)
                        else:
                            children_non_values.append((sv_eid, v))

                    for tgi, vals in children_values.items():
                        stmt_id = SGStatementNode.get_id(sgi, p, tgi)
                        if stmt_id not in sg.nodes:
                            sg.add_node(stmt_id, data=SGStatementNode.new(stmt_id))

                        if (sgi, stmt_id, p) not in sg.edges:
                            sg.add_edge(sgi, stmt_id, key=p, data=SGEdge(sgi, stmt_id, p, {}))
                        if (stmt_id, tgi, p) not in sg.edges:
                            sg.add_edge(stmt_id, tgi, key=p, data=SGEdge(stmt_id, tgi, p, {}))

                        sg_stmt: SGStatementNode = sg.nodes[stmt_id]['data']
                        for v in vals:
                            sg_stmt.track_provenance(SGEdgeFlowSource(uid, sgi, p), stmt.id, SGEdgeFlowTarget(v.id, tgi, p), stmt.get_provenance(EdgeFlowSource(uid, p), EdgeFlowTarget(v.id, p)))

                        for qp, qv in children_non_values:
                            qtgi = self.get_sg_node_id(qv)
                            if (stmt_id, qtgi, qp) not in dg.edges:
                                sg.add_edge(stmt_id, qtgi, key=qp, data=SGEdge(stmt_id, qtgi, qp, {}))

                            sg_stmt.track_provenance(SGEdgeFlowSource(uid, sgi, p), stmt.id, SGEdgeFlowTarget(qv.id, qtgi, qp), stmt.get_provenance(EdgeFlowSource(uid, p), EdgeFlowTarget(qv.id, qp)))
        args.sg = sg

    def cta_majority_vote(self, args: SemanticGraphConstructorArgs):
        # temporary before we should refactor the whole s*
        """Predict the column types for each column in the semantic graph"""
        def add_merge_qnodes(cell: CellNode, cell2qnodes: Dict[str, List[QNode]]):
            # attempt to merge qnodes (spatial) if they are contained in each other
            # we should go even higher order
            if len(cell.qnode_ids) > 1:
                # attempt to merge qnodes (spatial) if they are contained in each other
                # we should go even higher order
                ignore_qnodes = set()
                for q0_id in cell.qnode_ids:
                    q0 = self.qnodes[q0_id]
                    # location or located in the administrative entity
                    vals = {
                        stmt.value.as_qnode_id()
                        for stmt in chain(q0.props.get("P131", []), q0.props.get("P276", []))
                    }
                    for q1_id in cell.qnode_ids:
                        if q0_id == q1_id:
                            continue
                        if q1_id in vals:
                            # q0 is inside q1, ignore q1
                            ignore_qnodes.add(q1_id)
                cell2qnodes[cell.id] = [self.qnodes[q_id] for q_id in cell.qnode_ids if q_id not in ignore_qnodes]
            elif len(cell.qnode_ids) > 0:
                cell2qnodes[cell.id] = [self.qnodes[cell.qnode_ids[0]]]
            else:
                cell2qnodes[cell.id] = []

        args.cta = {}
        sg = args.sg
        dg = args.dg
        cell2qnodes = {}
        for uid, udata in sg.nodes(data=True):
            u: SGColumnNode = udata['data']
            if not u.is_column:
                continue

            # cells in this column
            cells: List[CellNode] = [dg.nodes[cid]['data'] for cid in u.nodes]
            covered_fractions = [
                sum(span.length for spans in cell.qnodes_span.values() for span in spans) / max(len(cell.value), 1)
                for cell in cells
                if len(cell.qnode_ids) >= 1
            ]
            if len(covered_fractions) == 0:
                continue
            avg_covered_fractions = np.mean(covered_fractions)
            if avg_covered_fractions < 0.8:
                continue

            for cell in cells:
                add_merge_qnodes(cell, cell2qnodes)
            merged_class_freq = dict(Counter([
                c
                for cell in cells
                for c in (stmt.value.as_qnode_id()
                    for qnode in cell2qnodes[cell.id]
                    for stmt in qnode.props.get("P31", [])
                )
            ]).items())
            if len(merged_class_freq) == 0:
                continue

            best_class, count = max(merged_class_freq.items(), key=itemgetter(1))
            args.cta[str(u.column)] = WDOnt.get_qnode_uri(best_class)
        
        return args

    def construct_sm(self, args: SemanticGraphConstructorArgs):
        """Get the final semantic model by combining relationships in SG and CTA"""
        sm = O.SemanticModel()
        readable_label_fn = lambda qnode_id: get_label(qnode_id, self.qnodes, self.wdclasses, self.wdprops)
        # a mapping from data node to class node for each data node that has a type
        dnode2cnode = {}
        for uid, udata in args.sg.nodes(data=True):
            u: SGNode = udata['data']
            if u.is_column:
                sm.add_node(O.DataNode(u.id, u.column, u.label))
                if u.column in args.cta:
                    qnode = self.wdclasses[args.cta[u.column]]
                    _cid = str(uuid4())
                    sm.add_node(O.ClassNode(_cid, qnode.get_uri(), f"wd:{qnode.id}", readable_label=readable_label_fn(qnode.id)))
                    dnode2cnode[u.id] = _cid
                    sm.add_edge(O.Edge(_cid, u.id, str(RDFS.label), "rdfs:label"))
            elif u.is_statement:
                sm.add_node(O.ClassNode(u.id, self.STATEMENT_URI, self.STATEMENT_REL_URI))
            elif u.is_value:
                if u.is_literal_value:
                    sm.add_node(O.LiteralNode(u.id, u.value.to_string_repr()))
                else:
                    sm.add_node(O.LiteralNode(u.id, u.value.as_qnode_id(), readable_label=readable_label_fn(u.value.as_qnode_id())))

        for uid, vid, eid, edata in args.sg.edges(data=True, keys=True):
            if uid in dnode2cnode:
                uid = dnode2cnode[uid]
            if vid in dnode2cnode:
                vid = dnode2cnode[vid]
            porq = self.wdprops[eid]
            sm.add_edge(O.Edge(uid, vid, porq.get_uri(), f"p:{porq.id}", readable_label=readable_label_fn(porq.id)))
        args.sm = sm

    def prune_sg_redundant_entity(self, args: SemanticGraphConstructorArgs):
        """Remove entity & context that is:
        - just an intermediate nodes that the parents only has one edge and the children has only one edge too
        - don't remove entity what is the main value of the statement
        """
        sm_g = args.sg

        # we are going to remove entity & context that is:
        # just a intermediate nodes that the parents only one edges and the children too
        stop = False
        while not stop:
            stop = True
            # for nid, ndata in list(sm_g.nodes(data=True)):
            #     n: SGNode = ndata['data']
            #     if n.is_statement:
            #         inedges: List[NxSGEdge] = list(sm_g.in_edges(nid, data=True, keys=True))
            #         outedges: List[NxSGEdge] = list(sm_g.out_edges(nid, data=True, keys=True))
            #
            #         if len(inedges) != 1:
            #             continue
            #         uid, _, seid, sedata = inedges[0]
            #         max_weight = -1
            #         for _, vid, teid, tedata in outedges:
            #             v: SGNode = sm_g.nodes[vid]['data']
            #             if teid != seid:
            #                 # we only want to consider value of the statement
            #                 continue
            #
            #             if v.is_column:
            #                 if tedata['data'].features['freq'] > max_weight:
            #                     max_weight = tedata['data'].features['freq']
            #         for _, vid, teid, tedata in outedges:
            #             v: SGNode = sm_g.nodes[vid]['data']
            #             if teid != seid:
            #                 # we only want to consider value of the statement
            #                 continue
            #
            #             if v.is_value:
            #                 if tedata['data'].features['freq'] < max_weight:
            #                     assert False
            #                     sm_g.remove_node(vid)
            #                     stop = False

            for nid, ndata in list(sm_g.nodes(data=True)):
                n: SGNode = ndata['data']
                if n.is_value and n.is_entity_value:
                    inedges: List[NxSGEdge] = list(sm_g.in_edges(nid, data=True, keys=True))
                    outedges: List[NxSGEdge] = list(sm_g.out_edges(nid, data=True, keys=True))

                    if len(inedges) == 0:
                        if all(sm_g.out_degree(vid) == 1 for _, vid, teid, tedata in outedges):
                            sm_g.remove_node(nid)
                            stop = False
                            continue

                    """
                    Remove the entity when it is not doing 
                    """
                    if len(inedges) > 1 or len(outedges) > 1:
                        continue

                    if len(inedges) > 0:
                        uid, _, seid, sedata = inedges[0]
                        if not all(sm_g.nodes[sib_nid]['data'].is_value for
                                   _, sib_nid in sm_g.out_edges(uid)):
                            continue

                    if len(outedges) > 0:
                        _, vid, teid, tedata = outedges[0]
                        if sm_g.out_degree(vid) != 1:
                            continue

                    # remove that node has only
                    sm_g.remove_node(nid)
                    stop = False

            for nid, ndata in list(sm_g.nodes(data=True)):
                n = ndata['data']
                if n.is_statement and (sm_g.in_degree(nid) == 0 or sm_g.out_degree(nid) == 0):
                    sm_g.remove_node(nid)
                    stop = False

    # ##################################################################################################################
    # CODE FOR COMPUTING FEATURES
    # ##################################################################################################################
    def calculate_link_frequency(self, args: SemanticGraphConstructorArgs):
        def get_in_edges(g, uid) -> List[SGEdge]:
            return [edata['data'] for _, _, eid, edata in g.in_edges(uid, data=True, keys=True)]

        sg = args.sg
        for sid, sdata in sg.nodes(data=True):
            s: SGStatementNode = sdata['data']
            if not s.is_statement:
                continue

            in_edge, = get_in_edges(sg, sid)
            in_edge.features['freq'] = s.compute_freq(sg, args.dg, in_edge, is_unique_freq=False)
            in_edge.features['unique_freq'] = s.compute_freq(sg, args.dg, in_edge, is_unique_freq=True)

            for _, vid, sv_eid, sv_edata in sg.out_edges(sid, data=True, keys=True):
                # count the frequency of the sv_eid by counting the number of times it is between specific instance
                sv_e: SGEdge = sv_edata['data']
                sv_e.features['freq'] = s.compute_freq(sg, args.dg, sv_e, is_unique_freq=False)
                sv_e.features['unique_freq'] = s.compute_freq(sg, args.dg, sv_e, is_unique_freq=True)

    # ##################################################################################################################
    # CODE FOR SOLVING THE STEINER TREE PROBLEM
    # ##################################################################################################################
    def st_maximum_arborescence_solver(self, args: SemanticGraphConstructorArgs):
        sm_g = nx.MultiDiGraph()
        for uid, vid, eid, edata in args.sg.edges(data=True, keys=True):
            if args.sg.nodes[vid]['data'].is_statement:
                # freq. for source -> statement
                # use a constant value as this link depends on the link coming out from the statement
                edata['weight'] = 0.1
            else:
                edata['weight'] = edata['data'].features['freq']

        # find the tree
        resp = nx.algorithms.tree.branchings.maximum_branching(args.sg, attr='weight', preserve_attrs=True)
        for s, t, e, edata in resp.edges(data=True, keys=True):
            edge = edata['data']
            if s not in sm_g:
                sm_g.add_node(s, **args.sg.nodes[s])
            if t not in sm_g:
                sm_g.add_node(t, **args.sg.nodes[t])
            sm_g.add_edge(s, t, key=edge.predicate, data=edge)

        # remove redundant nodes that is not relevant to the terminals
        # expand the statement with
        # for nid, ndata in list(sm_g.nodes(data=True)):
        #     n: SGNode = ndata['data']
        #     if n.is_statement:
        #         edges = list(sm_g.in_edges(nid, data=True, keys=True))
        #         if len(edges) == 0:
        #             continue
        #         assert len(edges) == 1
        #         uid, _, seid, edata = edges[0]
        #         source_e: SGEdge = edata['data']
        #
        #         # split when all props of statement is the same as the value of the statement
        #         need_split_stmt = True
        #         for _, vid, _, edata in sm_g.out_edges(nid, data=True, keys=True):
        #             if edata['data'].predicate != source_e.predicate:
        #                 need_split_stmt = False
        #                 break
        #
        #         if need_split_stmt:
        #             lst = list(sm_g.out_edges(nid, data=True, keys=True))
        #             for index, (_, vid, teid, edata) in enumerate(lst[1:]):
        #                 n2 = n.clone(without_provenance=True)
        #                 n2.id = n.id + f":{index}"
        #                 sm_g.add_node(n2.id, data=n2)
        #                 sm_g.add_edge(uid, n2.id, key=seid, data=source_e)
        #                 sm_g.add_edge(n2.id, vid, key=teid, data=edata['data'])
        #                 sm_g.remove_edge(n.id, vid, teid)

        # remove redundant nodes:
        # - a statement doesn't have any properties or incoming nodes
        # - an entity node that is a leaf node, and is the only child of the parent statement.
        while True:
            remove_nodes = set()
            for nid in sm_g:
                n: SGNode = sm_g.nodes[nid]['data']
                outdegree = sm_g.out_degree(nid)
                indegree = sm_g.in_degree(nid)

                if n.is_statement and (outdegree == 0 or indegree == 0):
                    # a statement doesn't have any properties or incoming nodes
                    remove_nodes.add(nid)
                elif n.is_value:
                    if outdegree == 0:
                        # an entity that doesn't have any outgoing edges, that can be used as something for the statement
                        # so we only remove it when the parent class has only one class
                        total_outdegree_parent = sum(
                            sm_g.out_degree(uid) for uid, vid, edata in sm_g.in_edges(nid, data=True))
                        if total_outdegree_parent == 1:
                            remove_nodes.add(nid)

                        # this extra entity will only be useful if it is not the qualifiers of the statement
                        if all(sm_g.in_degree(uid) == 1 and eid !=
                               next(iter(sm_g.in_edges(uid, data=True, keys=True)))[2] for uid, vid, eid, edata in
                               sm_g.in_edges(nid, data=True, keys=True)):
                            remove_nodes.add(nid)
                    if indegree == 0:
                        # an entity that doesn't have any incoming edges, but the only on predicate
                        total_outdegree_children = sum(
                            sm_g.out_degree(uid) for uid, vid, eid in sm_g.out_edges(nid, data=True))
                        if outdegree == 1 and total_outdegree_children == 1:
                            remove_nodes.add(nid)

            for nid in remove_nodes:
                sm_g.remove_node(nid)

            if len(remove_nodes) == 0:
                break

        # remove statement that has no qualifiers
        # for nid, ndata in list(sm_g.nodes(data=True)):
        #     n: SGNode = ndata['data']
        #     if n.is_statement and sm_g.in_degree(nid) == 1 and sm_g.out_degree(nid) == 1:
        #         uid, _, seid, sedata = list(sm_g.in_edges(nid, data=True, keys=True))[0]
        #         _, vid, teid, tedata = list(sm_g.out_edges(nid, data=True, keys=True))[0]
        #
        #         if sedata['data'].predicate == tedata['data'].predicate:
        #             # remove the statement that has no qualifiers
        #             sm_g.remove_node(nid)
        #             e: SGEdge = sedata['data'].clone(without_provenance=True)
        #             sm_g.add_edge(uid, vid, seid, data=e)
        args.sg = sm_g

    def st_networkx_st_solver(self, args: SemanticGraphConstructorArgs):
        sm_g = nx.MultiDiGraph()
        terminal_nodes = []
        for nid, ndata in args.sg.nodes(data=True):
            n: SGNode = ndata['data']
            if n.is_column or n.is_context:
                terminal_nodes.append(nid)

        return nx.algorithms.approximation.steinertree.steiner_tree(args.sg, terminal_nodes, weight='weight')


def viz_sg(sg, qnodes: Dict[str, QNode], wdclasses: Dict[str, WDClass], wdprops: Dict[str, WDProperty], outdir: str, graph_id: str):
    get_label_fn = lambda x: get_label(x, qnodes, wdclasses, wdprops)
    colors = {
        "context": dict(fill="#C6E5FF", stroke="#5B8FF9"),
        "statement": dict(fill="#d9d9d9", stroke="#434343"),
        "kg": dict(fill="#b7eb8f", stroke="#135200"),
        "column": dict(fill='#ffd666', stroke='#874d00')
    }

    def node_fn(uid, udata):
        u: SGNode = udata['data']
        html = ""
        label = ""
        if u.is_value:
            nodetype = 'context' if u.is_in_context else 'kg'
            label = u.label
            if u.is_entity_value:
                qnodeid = u.qnode_id
                html = f"""<a href="http://www.wikidata.org/wiki/{qnodeid}" target="_blank">{get_label_fn(qnodeid)}</a>"""
        elif u.is_statement:
            nodetype = 'statement'
        else:
            assert u.is_column
            label = f"{u.label} ({u.column})"
            nodetype = 'column'

        return {
            "label": label,
            "style": colors[nodetype],
            "labelCfg": {
                "style": {
                    "fill": "black",
                    "background": {
                        "padding": [4, 4, 4, 4],
                        "radius": 3,
                        **colors[nodetype]
                    }
                }
            },
            "html": html
        }

    def edge_fn(eid, edata):
        edge: SGEdge = edata['data']
        label = get_label_fn(eid)
        if 'prob' in edge.features:
            label += f" p={edge.features['prob']:.6f}"

        html = f'<a href="http://www.wikidata.org/wiki/Property:{eid}" target="_blank">{get_label_fn(eid)}</a>'
        if len(edge.features) > 0:
            html_feat = []
            for feat, val in edge.features.items():
                html_feat.append(f"<li><span style='min-width: 100px'><b>{feat}:</b></span> {val:.2f}</li>")
            html += f"<ul>{''.join(html_feat)}</ul>"

        return {
            "label": label,
            "html": html
        }

    return viz_graph(sg, node_fn, edge_fn, outdir, graph_id)


if __name__ == '__main__':
    from sm_unk.config import HOME_DIR
    from sm_unk.dev.wikitable2wikidata.sxx_evaluation import get_input_data
    from sm_unk.dev.wikitable2wikidata.kg_index import KGObjectIndex

    table_index = 6

    max_n_hop = 2
    dataset_dir = HOME_DIR / "wikitable2wikidata/250tables"
    gold_models = get_input_data(dataset_dir, dataset_dir.name, only_curated=True)

    REDIS_CACHE_URL = 'redis://localhost:6379/8'
    qnodes = get_qnodes(dataset_dir, n_hop=max_n_hop + 1, test=True)
    wdclasses = WDClass.from_file(dataset_dir / "ontology", load_parent_closure=True)
    wdprops = WDProperty.from_file(load_parent_closure=True)

    kg_index_file = dataset_dir / "kg_index" / "object_index.2hop_transitive.pkl.gz"
    if kg_index_file.exists():
        kg_object_index = KGObjectIndex.deserialize(kg_index_file, verbose=True)
    else:
        index_qnode_ids = list(get_qnodes(dataset_dir, n_hop=1).keys())
        kg_object_index = KGObjectIndex.from_qnodes(index_qnode_ids, qnodes, wdprops,
                                                    n_hop=2,
                                                    verbose=True)
        kg_object_index.serialize(kg_index_file)

    table = gold_models[table_index][1]
    dg = build_data_graph(table, qnodes, wdprops, kg_object_index, max_n_hop=2)
    constructor = SemanticGraphConstructor([
        SemanticGraphConstructor.init_sg,
        # SemanticGraphConstructor.calculate_link_frequency,
        # SemanticGraphConstructor.prune_sg_redundant_entity,
    ], qnodes, wdclasses, wdprops)
    resp = constructor.run(table, dg, debug=False)